# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent import scope
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.policy import registry
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.utils import (
    copy_events,
    copy_tags_in_span,
    copy_tags_to_offset,
    get_properties,
    get_last_event_id,
    track_string,
)
from contrast.utils.decorators import fail_quietly

from functools import lru_cache

from .split_propagator import SplitPropagator


@lru_cache(maxsize=10)
def _get_regex_policy_node(method_name):
    patch_policy = registry.get_policy_by_name(f"re.Match.{method_name}")
    return patch_policy.propagator_nodes[0] if patch_policy.propagator_nodes else None


def _propagate_group_string(
    method_name, target, source_properties, span, preshift, retval
):
    policy_node = _get_regex_policy_node(method_name)
    if policy_node is None:
        return

    target_properties = copy_tags_in_span(target, source_properties, span)
    # This can be None if len(target) < 2
    if target_properties is None:
        return

    parent_ids = [get_last_event_id(source_properties)]

    copy_events(target_properties, source_properties)

    target_properties.build_event(
        policy_node,
        target,
        preshift.obj,
        retval,
        preshift.args,
        preshift.kwargs,
        parent_ids,
    )


@fail_quietly("Failed to propagate regex group")
def propagate_group(self_obj, target, *args):
    """
    Propagator for re.Match.group()

    self_obj:
        re.Match object
    target:
        str result of calling .group()
    """
    source_properties = get_properties(self_obj.string)
    if source_properties is None:
        return

    # If no args are given, we simply process the 0th group, which is the
    # match for the entire pattern.
    if not args:
        args = [0]

    # If fewer than two args are given, the result is a single string match.
    # Otherwise, the result is an iterable of strings.
    if len(args) < 2:
        target = [target]

    preshift = Preshift(self_obj.string, args, {})

    for string, group in zip(target, args):
        _propagate_group_string(
            "group", string, source_properties, self_obj.span(group), preshift, target
        )


@fail_quietly("Failed to propagate regex groups")
def propagate_groups(self_obj, target, *args):
    """
    Propagator for re.Match.groups()

    self_obj:
        re.Match object
    target:
        tuple of str objects; result of calling .group()
    """
    # If there were no groups, this will be an empty tuple
    if not target:
        return

    source_properties = get_properties(self_obj.string)
    if source_properties is None:
        return

    preshift = Preshift(self_obj.string, args, {})

    for string, span in zip(target, self_obj.regs[1:]):
        _propagate_group_string(
            "groups", string, source_properties, span, preshift, target
        )


@fail_quietly("Failed to propagate regex groupdict")
def propagate_groupdict(self_obj, target, *args):
    """
    Propagator for re.Match.groupdict()

    self_obj:
        re.Match object
    target:
        dict with group names as keys and matched strings as values
    """
    # If there were no named groups, this will be an empty tuple
    if not target:
        return

    source_properties = get_properties(self_obj.string)
    if source_properties is None:
        return

    preshift = Preshift(self_obj.string, args, {})

    for name, value in target.items():
        _propagate_group_string(
            "groupdict", value, source_properties, self_obj.span(name), preshift, target
        )


def _propagate_sub_tags(
    target_properties,
    orig_properties,
    repl_properties,
    string,
    count,
    source_matches,
    repl_results,
):
    source_offset = 0
    target_offset = 0

    for i in range(count):
        match = source_matches[i]
        # Copy any tags from the source string prior to the match
        if orig_properties is not None:
            span = AdjustedSpan(source_offset, match.start())
            source_tags = orig_properties.tags_at_range(span)
            copy_tags_to_offset(target_properties, source_tags, target_offset)

        target_offset += match.start() - source_offset
        source_offset = match.end()

        # Copy any tags from the replacement string
        if repl_properties[i] is not None:
            copy_tags_to_offset(
                target_properties, repl_properties[i].tags, target_offset
            )

        target_offset += len(repl_results[i])

    if source_offset < len(string) and orig_properties is not None:
        old_span = AdjustedSpan(source_offset, len(string))
        source_tags = orig_properties.tags_at_range(old_span)
        copy_tags_to_offset(target_properties, source_tags, target_offset)


def _propagate_sub_events(
    node,
    target_properties,
    orig_properties,
    repl_properties,
    target,
    args,
    kwargs,
):
    parent_ids = []
    if orig_properties is not None:
        parent_ids.append(get_last_event_id(orig_properties))

    copy_events(target_properties, orig_properties)

    for props in repl_properties:
        copy_events(target_properties, props)
        if props is not None:
            parent_ids.append(get_last_event_id(props))

    target_properties.build_event(node, target, None, target, args, kwargs, parent_ids)


@fail_quietly("Failed to propagate regex sub(n)")
def propagate_sub(node, target, repl_results, pattern, repl, string, count=0, **kwargs):
    """
    Propagator for re.sub and re.subn
    """
    with scope.contrast_scope():
        # contrast scope required here to prevent analysis in finditer patch.
        source_matches = (
            list(pattern.finditer(string))
            if node.instance_method
            else list(re.finditer(pattern, string, kwargs.get("flags", 0)))
        )

    # No propagation necessary because no replacement occurred
    if len(source_matches) == 0:
        return

    count = min(len(source_matches), count) if count > 0 else len(source_matches)

    # If the repl was not a callable, then the replacement string is the same
    # for each occurrence of the pattern
    if repl_results is None:
        repl_results = [repl] * len(source_matches)

    orig_properties = get_properties(string)
    repl_properties = [get_properties(x) for x in repl_results]
    if orig_properties is None and not any(repl_properties):
        return

    target_properties = get_properties(target) or track_string(target)
    if target_properties is None:
        # Nothing to do here since we can't copy tags and events to NoneType
        return

    _propagate_sub_tags(
        target_properties,
        orig_properties,
        repl_properties,
        string,
        count,
        source_matches,
        repl_results,
    )

    args = (pattern, repl, string)
    _propagate_sub_events(
        node,
        target_properties,
        orig_properties,
        repl_properties,
        target,
        args,
        kwargs,
    )

    target_properties.cleanup_tags()


class RegexSplitPropagator(SplitPropagator):
    @property
    def needs_propagation(self):
        if not self.preshift:
            return False

        return self.any_source_tracked

    @property
    def is_pattern_method(self):
        return self.node.class_name == "Pattern"

    def _get_flags(self):
        if self.is_pattern_method:
            return self.preshift.obj.flags
        if len(self.preshift.args) == 4:
            return self.preshift.args[-1]
        return self.preshift.kwargs.get("flags", 0)

    def _get_pattern(self):
        return self.preshift.obj if self.is_pattern_method else self.preshift.args[0]

    def _propagate(self, string, flags, pattern, source_properties):
        source_offset = 0
        target_index = 0

        while target_index < len(self.target):
            # If the pattern passed to re.split contains any groupings, then
            # those groupings need to be propagated appropriately.
            if self.is_pattern_method:
                match = pattern.match(string[source_offset:])
            else:
                match = re.match(pattern, string[source_offset:], flags=flags)
            if match is not None:
                for target, span in zip(self.target[target_index:], match.regs[1:]):
                    target_properties = copy_tags_in_span(
                        target, source_properties, span, offset=source_offset
                    )
                    copy_events(target_properties, source_properties)
                    target_index += 1
                source_offset += len(match.group())

            if target_index < len(self.target):
                target = self.target[target_index]
                if target is not None:
                    span = (source_offset, source_offset + len(target))
                    target_properties = copy_tags_in_span(
                        target, source_properties, span
                    )
                    copy_events(target_properties, source_properties)
                    source_offset += len(target)

                target_index += 1

    def propagate(self):
        string = self.sources[0]
        flags = self._get_flags()
        pattern = self._get_pattern()

        source_properties = get_properties(string)
        if source_properties is None:
            return

        self._propagate(string, flags, pattern, source_properties)


class RegexFindallPropagator(RegexSplitPropagator):
    def propagate(self):
        """
        Propagation action for re.findall and re.Pattern.findall

        When the number of groups in the given pattern is 1 or 0 (which indicates an
        implicit group corresponding to the whole pattern), findall returns a flat list
        representing one result per match. In this case, the propagation action is
        identical to the action for re.split, so we simply call the parent method.

        When the number of groups is greater than 1, findall returns a list of tuples
        where each tuple represents the groups corresponding to a given match. In this
        case, we use finditer to get the corresponding Match instances in order to
        determine the spans to use for propagation.
        """
        string = self.sources[0]
        source_props = get_properties(string)
        if source_props is None:
            return

        flags = self._get_flags()
        pattern = re.compile(self._get_pattern())
        if pattern.groups < 2:
            self._propagate(string, flags, pattern, source_props)
            return

        # We know we're in scope here, so we're safe to call finditer
        matches = (
            pattern.finditer(string)
            if self.is_pattern_method
            else re.finditer(pattern, string, flags)
        )

        for target, match in zip(self.target, matches):
            for i, result in enumerate(target):
                target_properties = copy_tags_in_span(
                    result,
                    source_props,
                    # skip the first span since it duplicates the entire match
                    match.span(i + 1),
                )

                # This can be None if len(target) < 2
                if target_properties is None:
                    continue

                parent_ids = [get_last_event_id(source_props)]

                copy_events(target_properties, source_props)

                target_properties.build_event(
                    self.node,
                    result,
                    self.preshift.obj,
                    result,
                    self.preshift.args,
                    self.preshift.kwargs,
                    parent_ids,
                )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Utilities for updating preflight hashes with information from requests

This module is primarily tested by framework tests.
"""
from contrast.utils.decorators import fail_loudly
from contrast.utils.digest_utils import Digest

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def _get_route_for_hash(context):
    """
    Get route to be used for hashing according to preflight spec:

        1. Route signature
        2. Normalized uri - the agents best attempt at normalizing the URI
           for path parameters. typically this involves a regex to abstract
           numeric or UUID path segments, but is not well defined. ideally,
           this case will phase out as route coverage expands, so the
           definition does not need to be consistent across agents so long
           as it is consistent within them.
        3. Raw uri
    """
    route = context.current_route
    if route is not None and route.signature:
        return route.signature

    normalized_uri = context.request.get_normalized_uri()
    return normalized_uri or context.request.path_qs


def _get_verb_for_hash(context):
    route = context.current_route
    if route is not None and route.verb:
        return route.verb

    return context.request.method


@fail_loudly("Failed to update preflight hashes")
def update_preflight_hashes(context):
    """
    Update preflight hashes for all findings with available route/request info

    @param context: Current request context
    """
    for finding in context.findings:
        hasher = Digest()
        hasher.update(finding.hash_code)

        hasher.update(_get_route_for_hash(context))
        hasher.update(_get_verb_for_hash(context))

        hash_code = hasher.finish()

        finding.hash_code = hash_code
        finding.preflight = ",".join([finding.rule_id, hash_code])
        logger.debug("updated preflight value: %s", finding.preflight)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor.wrapt import register_post_import_hook
import sys
from contrast.agent.policy import patch_manager
from contrast.patches.databases import dbapi2
from contrast.utils.decorators import fail_quietly

SQLITE3 = "sqlite3"
PYSQLITE2_DBAPI2 = "pysqlite2.dbapi2"
VENDOR = "SQLite3"


class Sqlite3Patcher(dbapi2.Dbapi2Patcher):
    @fail_quietly("failed to get database inventory information")
    def init_dbinfo(self, connection, connect_args, connect_kwargs):
        # sqlite does not use a server, so no need for host/port
        database = connect_kwargs.get("database") or (
            connect_args[0] if len(connect_args) > 0 else "unknown"
        )
        self.dbinfo["database"] = database


def instrument_sqlite3(sqlite3):
    dbapi2.instrument_adapter(
        sqlite3, VENDOR, Sqlite3Patcher, extra_cursors=[sqlite3.Cursor]
    )


def instrument_pysqlite2_dbapi2(pysqlite2_dbapi2):
    """Supports the older pysqlite module in Py2"""
    dbapi2.instrument_adapter(
        pysqlite2_dbapi2,
        VENDOR,
        Sqlite3Patcher,
        extra_cursors=[pysqlite2_dbapi2.Cursor],
    )


def register_patches():
    register_post_import_hook(instrument_sqlite3, SQLITE3)
    register_post_import_hook(instrument_pysqlite2_dbapi2, PYSQLITE2_DBAPI2)


def reverse_patches():
    sqlite3 = sys.modules.get(SQLITE3)
    if sqlite3:
        patch_manager.reverse_patches_by_owner(sqlite3)
        patch_manager.reverse_patches_by_owner(sqlite3.Cursor)
        patch_manager.reverse_patches_by_owner(sqlite3.Connection)

    psqlite = sys.modules.get(PYSQLITE2_DBAPI2)
    if psqlite:
        patch_manager.reverse_patches_by_owner(psqlite)
        patch_manager.reverse_patches_by_owner(psqlite.Cursor)
        patch_manager.reverse_patches_by_owner(psqlite.Connection)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage, PYTHON
from ._traces import _Traces
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Preflight(BaseTsAppMessage):
    def __init__(self, findings, request):
        super().__init__()

        self.findings = findings
        # NOTE: we cannot send our Request object to other threads, because its data is
        # invalidated by webob once the server sends the associated response. Any work
        # that needs to happen with the request object must be guaranteed to finish
        # before our middleware sends back a response.
        self.reportable_request = request and request.reportable_format

        self.body = {"messages": []}
        for idx, finding in enumerate(self.findings):
            message = {
                "appLanguage": PYTHON,
                "appName": self.settings.app_name,
                "appPath": self.settings.app_path,
                "appVersion": self.settings.app_version,
                "code": "TRACE",
                "data": finding.preflight,
                "key": idx,
            }
            self.body["messages"].append(message)

    @property
    def name(self):
        return "preflight"

    @property
    def path(self):
        return "preflight"

    @property
    def request_method(self):
        return requests.put

    @property
    def expected_response_codes(self):
        return [200]

    @fail_loudly("Failed to process Preflight response")
    def process_response(self, response, reporting_client):
        logger.debug("PREFLIGHT: Processing message")
        if not self.process_response_code(response, reporting_client):
            logger.debug("PREFLIGHT: invalid response code; exiting")
            return

        body = response.text
        finding_idxs_to_report = self._parse_body(body)
        if finding_idxs_to_report is not None and len(finding_idxs_to_report) > 0:
            logger.debug("PREFLIGHT: found indexes to report")
        else:
            logger.debug("PREFLIGHT: no indexes to report; exiting")
        for finding_idx in finding_idxs_to_report:
            finding = self.findings[finding_idx]
            logger.debug(
                "PREFLIGHT: generating Trace from finding for %s", finding.rule_id
            )
            traces_msg = _Traces(finding, self.reportable_request)
            logger.debug("PREFLIGHT: generated Trace from finding; queuing")
            reporting_client.add_message(traces_msg)
            logger.debug("PREFLIGHT: Trace queued")

    @staticmethod
    def _parse_body(body):
        """
        A preflight response body is a comma-separated list of finding indices that
        should be reported in a Traces message. Some elements of this list will have a
        *, meaning TS needs an AppCreate message before it will accept this finding. For
        now, we do not send findings with a *.

        TODO: PYT-2119 handle * preflight findings
        """
        indices = body.strip('"').split(",")
        return [int(index) for index in indices if index.isdigit()]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.triggers.trigger_config_rule import TriggerConfigRule


class SessionRewritingRule(TriggerConfigRule):
    @property
    def name(self):
        return "session-rewriting"

    def is_violated(self, node, source, **kwargs):
        """
        The rule is violated if the value is False or if it is not set at all (None)
        """
        if self.count_threshold_reached():
            return False
        return not source

    def is_violated_properties(*args, **kwargs):
        """Override for safety. In theory, this should never be called."""
        return False



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.policy.propagators import BasePropagator
from contrast.agent.assess.utils import copy_events, get_properties
from contrast.utils.assess.tag_utils import merge_tags


class SlicePropagator(BasePropagator):
    def _propagate(self):
        slice_ = self.preshift.args[0] if self.preshift.args else None

        # Offset in the target string
        target_offset = 0

        source_properties = get_properties(self.preshift.obj)

        # This is like the coolest thing ever. I found it here:
        # https://stackoverflow.com/a/42883770/4312739
        # It converts a slice to a range of indices
        offsets = list(range(self.preshift.obj_length)[slice_])

        for source_offset in offsets:
            span = AdjustedSpan(source_offset, source_offset + 1)
            source_tags = source_properties.tags_at_range(span)
            for name in source_tags:
                end_offset = target_offset + 1
                new_span = AdjustedSpan(target_offset, end_offset)
                self.target_properties.add_tag(name, new_span)

            target_offset += 1

        copy_events(self.target_properties, source_properties)
        merge_tags(self.target_properties.tags)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading

from contrast.agent.policy.constants import (
    ALL_ARGS,
    ALL_KWARGS,
    OBJECT,
    RETURN,
    TRIGGER_TYPE,
    CREATION_TYPE,
)

from contrast.agent.assess.truncate import truncate, truncate_tainted_string
from contrast.agent.assess.utils import get_properties
from contrast.agent.settings import Settings
from contrast.api.trace_event import (
    TraceEvent,
    TraceEventObject,
    ParentObjectId,
    TraceEventSource,
)
from contrast.utils.base64_utils import base64_encode
from contrast.utils.object_utils import safe_copy
from contrast.utils.stack_trace_utils import (
    build_stack,
    clean_stack,
)
from contrast.utils.string_utils import ensure_string
from contrast.utils.timer import now_ms
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

INIT = "__init__"
INITIALIZERS = (INIT, "__new__")
NONE_STRING = str(None)


class _TriggerTimeMixin(object):
    """
    Contrast Event methods only called on all ContrastEvent instances
    if a trigger is reached.
    """

    def to_reportable_event(self):
        """
        Convert a ContrastEvent to a TraceEvent.
        """
        self._update_init_return()

        event = TraceEvent()

        event.type = self.node.node_type
        event.action = self.event_action
        event.timestamp_ms = self.time
        event.thread = str(self.thread)
        event.tags = self.node.tags

        self._build_all_event_objects(event)

        event.stack = clean_stack(self._raw_stack, depth=20)

        safe_source_name = self.source_name if self.source_name is not None else ""
        event.field_name = safe_source_name

        if self.source_type:
            event.event_sources.extend([self.build_source_event(safe_source_name)])

        event.object_id = int(self.event_id)

        if self.parent_ids:
            for parent_id in self.parent_ids:
                parent = ParentObjectId()
                parent.id = parent_id
                event.parent_object_ids.extend([parent])

        self._build_complete_signature(event)
        self._validate_event(event)

        return event

    def _update_init_return(self):
        """
        For purposes of pretty reporting in Teamserver, we will say the
        `__init__` instance method return the `self` object (the object getting
        instantiated) instead of `None`, even though `None` is the return value of
        `__init__` methods.

        This will not apply if the node is not a class (it's possible someone
        creates a module level function called `__init__` or if the return
        value is already populated (for safety).
        """
        if self.node.method_name == INIT and self.node.class_name is not None:
            self.ret = self.ret or self.obj

    def build_source_event(self, safe_source_name):
        """
        Create a new TraceEventSource

        :param safe_source_name: source name or empty string
        :return: TraceEvenSource
        """
        trace_event_source = TraceEventSource()

        trace_event_source.type = self.source_type
        trace_event_source.name = safe_source_name

        return trace_event_source

    def _set_event_source_and_target(self, event):
        """
        We have to do a little work to figure out what our TS appropriate
        target is. To break this down, the logic is as follows:
        1) If my node has a target, work on targets. Else, work on sources.
           Per TS law, each node must have at least a source or a target.
           The only type of node w/o targets is a Trigger, but that may
           change.
        2) I'll set the event's source and target to TS values.
        """
        if self.node.targets:
            event.source = self.node.ts_valid_source
            event.target = self.node.ts_valid_target
        elif self.node.sources:
            event.source = self.node.ts_valid_target or self.node.ts_valid_source

    def _build_all_event_objects(self, event):
        """
        First populate event.source and event.target
        Then populate fields of event.object, event.ret, and event.args (contains both
        args and kwargs for python) which are each `TraceEventObject`s
        """
        self._set_event_source_and_target(event)

        self._build_event_object(
            event, event.object, self.obj, self.taint_location == OBJECT
        )
        self._build_event_object(
            event, event.ret, self.ret, self.taint_location == RETURN
        )

        for index, arg in enumerate(self.args):
            event_obj = TraceEventObject()
            self._build_event_object(
                event, event_obj, arg, self.taint_location == index
            )
            event.args.append(event_obj)

        if not self.kwargs:
            return

        event_obj = TraceEventObject()
        is_kwargs_taint_target = self.taint_location is ALL_KWARGS or isinstance(
            self.taint_location, str
        )
        self._build_event_object(event, event_obj, self.kwargs, is_kwargs_taint_target)
        event.args.append(event_obj)

    def _build_event_object(self, event, event_obj, obj, is_taint_target):
        str_representation, splat = _obj_to_str(obj)

        splat_len = None
        if splat:
            splat_len = len(str_representation)

        if is_taint_target:
            self._add_taint_ranges(event, event_obj, obj, splat_len)
            str_representation = truncate_tainted_string(str_representation, event)
        else:
            str_representation = truncate(str_representation)

        event_obj.value = base64_encode(str_representation)

    def _add_taint_ranges(self, event, event_obj, target, splat_len):
        """
        Populate event.taint_ranges
        """
        if self.taint_location is None:
            return

        if isinstance(target, dict):
            if self.taint_location == ALL_KWARGS and self.possible_key:
                properties = get_properties(target.get(self.possible_key, None))
            else:
                properties = get_properties(target.get(self.taint_location, None))
        else:
            properties = get_properties(target)

        if properties is None:
            return

        if splat_len is not None:
            tag_ranges = properties.tags_to_ts_obj(splat_range=(0, splat_len))
        else:
            tag_ranges = properties.tags_to_ts_obj()

        event.taint_ranges.extend(tag_ranges)
        # For now, only the taint_target needs to be officially marked as tracked.
        # This means that agent-tagged strings may not be marked as "tracked" for TS.
        # This may change in the future if we change the corresponding TS endpoint; in
        # that case, use recursive_is_tracked for each TraceEventObject.
        event_obj.tracked = True

    def _build_complete_signature(self, event):
        return_type = type(self.ret).__name__ if self.ret is not None else NONE_STRING

        event.signature.return_type = return_type
        # We don't want to report "BUILTIN" as a module name in Team Server
        event.signature.class_name = self.node.location.replace("BUILTIN.", "")
        event.signature.method_name = self.node.method_name

        if self.args:
            for item in self.args:
                arg_type = type(item).__name__ if item else NONE_STRING
                event.signature.arg_types.append(arg_type)

        if self.kwargs:
            arg_type = type(self.kwargs).__name__
            event.signature.arg_types.append(arg_type)

        event.signature.constructor = self.node.method_name in INITIALIZERS

        # python always returns None if not returned
        event.signature.void_method = False

        if not self.node.instance_method:
            event.signature.flags = 8

    def _validate_event(self, event):
        """
        TS is not able to render a vulnerability correctly if the source string index 0
        of the trigger event, ie event.source, is not a known one.

        See TS repo DataFlowSnippetBuilderVersion1.java:buildMarkup

        :param event: TraceEvent
        :return: None
        """
        allowed_trigger_sources = ["O", "P", "R"]
        if (
            event.action == TraceEvent.Action[TRIGGER_TYPE]
            and event.source[0] not in allowed_trigger_sources
        ):
            # If this is logged, check the node in policy.json corresponding to
            # this event and how the agent has transformed the source string
            logger.debug("WARNING: trigger event TS-invalid source %s", event.source)


class ContrastEvent(_TriggerTimeMixin):
    """
    This class holds the data about an event in the application
    We'll use it to build an event that TeamServer can consume if
    the object to which this event belongs ends in a trigger.
    """

    ATOMIC_ID = 0

    def __init__(
        self,
        node,
        tagged,
        self_obj,
        ret,
        args,
        kwargs,
        parent_ids,
        possible_key,
        source_type=None,
        source_name=None,
    ):
        self.node = node
        self.tagged = tagged
        self.source_type = source_type
        self.source_name = source_name
        self.parent_ids = parent_ids
        self.possible_key = possible_key
        self.obj = self_obj
        self.ret = ret
        self.args = safe_copy(args)
        self.kwargs = safe_copy(kwargs)

        # These are needed only at trigger-time but values must be set at init.
        self.time = now_ms()
        self.thread = threading.current_thread().ident
        self.event_id = ContrastEvent._atomic_id()

        self.event_action = self.node.build_action()

        self._raw_stack = build_stack() if self._should_get_stack else []

        # This must happen at init for stream events to work.
        self.taint_location = None
        self._update_method_information()

    @property
    def _should_get_stack(self):
        """
        Determine if event.stack should be populated or not.

        Get stacktrace for the event EXCEPT if
          1. the node explicitly indicates not to collect a stacktrace, which is the
                case for WSGI environ sources
          2. assess.stacktraces is configured to NONE
          3. assess.stacktraces is configured to SOME and event action is not
                creation or trigger


        :return: bool if to get stacktrace
        """
        settings = Settings()

        reportable_actions = (
            TraceEvent.Action[CREATION_TYPE],
            TraceEvent.Action[TRIGGER_TYPE],
        )

        if (
            self.node.skip_stacktrace
            or settings.is_collect_stacktraces_none()
            or (
                settings.is_collect_stacktraces_some()
                and self.event_action not in reportable_actions
            )
        ):
            return False

        return True

    @classmethod
    def _atomic_id(cls):
        ret = cls.ATOMIC_ID
        cls.ATOMIC_ID += 1
        return ret

    def _update_method_information(self):
        """
        For nicer reporting, we lie about the tagged value. For example, a call to
        split() returns a list of strings: ["foo", "bar"]. In the properties for "foo",
        the split event shows a return value of only "foo" instead of the whole list.
        """
        self.taint_location = self._find_taint_location()

        if self.taint_location is None:
            # This would be for trigger nodes without source or target. Trigger rule was
            # violated simply by a method being called. We'll save all the information,
            # but nothing will be marked up, as nothing need be tracked.
            return

        if self.taint_location == OBJECT:
            self.obj = self.tagged
            return

        if self.taint_location == RETURN:
            self.ret = self.tagged

    def _find_taint_location(self):
        """
        Based on what we know about the call that caused this event's creation,
        determine the appropriate taint location. Fall back to returning the
        first element in candidate_taint_locations. This method is greedy, so it
        will return the first valid taint location based on the order of candidate
        locations provided.

        Example:

        args = ("irrelevant string",)
        kwargs = {"irrelevant_key": "irrelevant_value", "important_key": "user_input"}
        candidate_taint_locations = [1, 2, "important_key"]

        In this case, we return "important_key" since there is no ARG_1 or ARG_2.

        @return: The appropriate element of candidate_taint_locations, or the first
            element if no suitable matches are found, or None if the list is empty.
        """
        candidate_taint_locations = self.node.targets or self.node.sources

        if not candidate_taint_locations:
            return None

        for candidate_location in candidate_taint_locations:
            found = (
                (candidate_location == RETURN)
                or (candidate_location == OBJECT and self.obj is not None)
                or (candidate_location == ALL_ARGS and self.args)
                or (candidate_location == ALL_KWARGS and self.kwargs)
                or (
                    isinstance(candidate_location, int)
                    and candidate_location < len(self.args)
                )
                or (
                    isinstance(candidate_location, str)
                    and candidate_location in self.kwargs
                )
            )
            # pylint doesn't like this in an if statement, but an assignment is ok
            if found:
                return candidate_location

        logger.debug(
            "WARNING: unable to find event taint location: %s %s %s %s %s",
            self.obj,
            self.ret,
            self.args,
            self.kwargs,
            candidate_taint_locations,
        )
        return candidate_taint_locations[0]


def _obj_to_str(self_obj):
    """
    Attempt to get a string representation of an object

    Right now we do our best to decode the object, but we handle any
    decoding errors by replacing with �. This technically is a loss
    of information when presented in TS, but it allows us to preserve
    the taint range information, which arguably is more important for
    Assess. In the future we might want to implement more robust
    handling of non-decodable binary data (i.e. to display escaped
    data with an updated taint range).

    If the object isn't stringy, then just return the string
    representation. In this case, we will need to splat the displayed
    taint range since we're not able to map tag ranges.

    :param self_obj: any python object, str, byte, bytearray, etc
    :return:
        1. str representing the object
        2. whether to splat the taint ranges or not, depending on if we can stringify
           the obj successfully
    """
    splat = False

    try:
        if isinstance(self_obj, bytearray):
            str_representation = self_obj.decode(errors="replace")
        else:
            str_representation = ensure_string(self_obj, errors="replace")
        if len(str_representation) != len(self_obj):
            # tag ranges won't line up properly if the length of the string
            # representation differs from the original length
            splat = True
    except Exception:
        str_representation = str(self_obj)
        splat = True

    return str_representation, splat



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from pip._vendor import pkg_resources
from collections import namedtuple

from contrast.agent import scope
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

Version = namedtuple("Version", ["major", "minor", "patch"])
DEFAULT_FRAMEWORK = "wsgi"
SUPPORTED_SERVERS = ["uwsgi", "gunicorn", "uvicorn", "mod_wsgi"]
DEFAULT_SERVER = "Unknown"


class DiscoverablePackage(object):
    def __init__(self, framework_name, default_package):
        self._name = ""
        self.version = None
        self.packages = framework_name if framework_name else DEFAULT_FRAMEWORK
        self.default_package = default_package
        self.set_info()

    @property
    def name(self):
        return self._name.capitalize()

    @property
    def name_lower(self):
        return self._name.lower()

    @property
    def full_version(self):
        return f"{self.version.major}.{self.version.minor}.{self.version.patch}"

    def set_info(self):
        raise NotImplementedError("Must implement set_info")

    def discover_framework(self, framework_name):
        """
        Except in the agent's own testing environment, the assumption here is
        that all environments using the agent will have only
        one supported framework or server.

        :return pkg_resources.DistInfoDistribution instance
        """
        with scope.contrast_scope():
            # Enter scope to prevent a recursive bug in pkg_resources which
            # calls on our re_patch and attempts to re-initialize this class.
            try:
                return pkg_resources.get_distribution(framework_name)
            except Exception:
                return None

    def discover_server(self):
        """
        Except in the agent's own testing environment, the assumption here is
        that all environments using the agent will have only
        one supported framework or server.
        :return pkg_resources.DistInfoDistribution instance
        """
        with scope.contrast_scope():
            # Enter scope to prevent a recursive bug in pkg_resources which
            # calls on our re_patch and attempts to re-initialize this class.
            for framework_name in self.packages:
                try:
                    return pkg_resources.get_distribution(framework_name)
                except Exception:
                    continue

        return None

    def __repr__(self):
        return f"{self.name} {self.full_version}"


class Framework(DiscoverablePackage):
    """
    A class to store information about the current web framework used in an application
    """

    def __init__(self, framework_name):
        super().__init__(framework_name, DEFAULT_FRAMEWORK)

    def set_info(self):
        framework = (
            self.discover_framework(self.packages) if self.packages != "wsgi" else None
        )
        if framework:
            version = framework.version.split(".")
            patch = version[2] if len(version) > 2 else "0"
            self.version = Version(major=version[0], minor=version[1], patch=patch)
            self._name = framework.project_name
        else:
            self._name = "wsgi"
            self.version = Version(major="0", minor="0", patch="0")


class Server(DiscoverablePackage):
    """
    A class to store information about the current web server used in an application
    """

    def __init__(self):
        super().__init__(SUPPORTED_SERVERS, DEFAULT_SERVER)

    def set_info(self):
        framework = self.discover_server()
        if framework:
            version = framework.version.split(".")
            patch = version[2] if len(version) > 2 else "0"
            self.version = Version(major=version[0], minor=version[1], patch=patch)
            self._name = framework.project_name
        else:
            logger.debug(
                "Did not find the current %s. Assuming it's %s.",
                self.__class__.__name__,
                self.default_package,
            )
            self._name = self.default_package
            self.version = Version(major="0", minor="0", patch="0")



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Logger classes responsible for output.
"""

from __future__ import annotations

import copy
import sys
import threading

from pickle import PicklingError
from sys import stderr, stdout
from typing import IO, Any, BinaryIO, TextIO

from contrast_vendor.structlog._utils import until_not_interrupted


WRITE_LOCKS: dict[IO[Any], threading.Lock] = {}


def _get_lock_for_file(file: IO[Any]) -> threading.Lock:
    global WRITE_LOCKS

    lock = WRITE_LOCKS.get(file)
    if lock is None:
        lock = threading.Lock()
        WRITE_LOCKS[file] = lock

    return lock


class PrintLogger:
    """
    Print events into a file.

    :param file: File to print to. (default: `sys.stdout`)

    >>> from contrast_vendor.structlog import PrintLogger
    >>> PrintLogger().info("hello")
    hello

    Useful if you follow
    `current logging best practices <logging-best-practices>`.

    Also very useful for testing and examples since `logging` is finicky in
    doctests.

    .. versionchanged:: 22.1
       The implementation has been switched to use `print` for better
       monkeypatchability.
    """

    def __init__(self, file: TextIO | None = None):
        self._file = file or stdout

        self._lock = _get_lock_for_file(self._file)

    def __getstate__(self) -> str:
        """
        Our __getattr__ magic makes this necessary.
        """
        if self._file is stdout:
            return "stdout"

        if self._file is stderr:
            return "stderr"

        raise PicklingError(
            "Only PrintLoggers to sys.stdout and sys.stderr can be pickled."
        )

    def __setstate__(self, state: Any) -> None:
        """
        Our __getattr__ magic makes this necessary.
        """
        if state == "stdout":
            self._file = stdout
        else:
            self._file = stderr

        self._lock = _get_lock_for_file(self._file)

    def __deepcopy__(self, memodict: dict[Any, Any] = {}) -> PrintLogger:
        """
        Create a new PrintLogger with the same attributes. Similar to pickling.
        """
        if self._file not in (stdout, stderr):
            raise copy.error(
                "Only PrintLoggers to sys.stdout and sys.stderr "
                "can be deepcopied."
            )

        newself = self.__class__(self._file)

        newself._lock = _get_lock_for_file(newself._file)

        return newself

    def __repr__(self) -> str:
        return f"<PrintLogger(file={self._file!r})>"

    def msg(self, message: str) -> None:
        """
        Print *message*.
        """
        f = self._file if self._file is not stdout else None
        with self._lock:
            until_not_interrupted(print, message, file=f, flush=True)

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg


class PrintLoggerFactory:
    r"""
    Produce `PrintLogger`\ s.

    To be used with `structlog.configure`\ 's ``logger_factory``.

    :param file: File to print to. (default: `sys.stdout`)

    Positional arguments are silently ignored.

    .. versionadded:: 0.4.0
    """

    def __init__(self, file: TextIO | None = None):
        self._file = file

    def __call__(self, *args: Any) -> PrintLogger:
        return PrintLogger(self._file)


class WriteLogger:
    """
    Write events into a file.

    :param file: File to print to. (default: `sys.stdout`)

    >>> from contrast_vendor.structlog import WriteLogger
    >>> WriteLogger().info("hello")
    hello

    Useful if you follow
    `current logging best practices <logging-best-practices>`.

    Also very useful for testing and examples since `logging` is finicky in
    doctests.

    A little faster and a little less versatile than `structlog.PrintLogger`.

    .. versionadded:: 22.1
    """

    def __init__(self, file: TextIO | None = None):
        self._file = file or sys.stdout
        self._write = self._file.write
        self._flush = self._file.flush

        self._lock = _get_lock_for_file(self._file)

    def __getstate__(self) -> str:
        """
        Our __getattr__ magic makes this necessary.
        """
        if self._file is stdout:
            return "stdout"

        if self._file is stderr:
            return "stderr"

        raise PicklingError(
            "Only WriteLoggers to sys.stdout and sys.stderr can be pickled."
        )

    def __setstate__(self, state: Any) -> None:
        """
        Our __getattr__ magic makes this necessary.
        """
        if state == "stdout":
            self._file = stdout
        else:
            self._file = stderr

        self._lock = _get_lock_for_file(self._file)

    def __deepcopy__(self, memodict: dict[Any, Any] = {}) -> WriteLogger:
        """
        Create a new WriteLogger with the same attributes. Similar to pickling.
        """
        if self._file not in (sys.stdout, sys.stderr):
            raise copy.error(
                "Only WriteLoggers to sys.stdout and sys.stderr "
                "can be deepcopied."
            )

        newself = self.__class__(self._file)

        newself._write = newself._file.write
        newself._flush = newself._file.flush
        newself._lock = _get_lock_for_file(newself._file)

        return newself

    def __repr__(self) -> str:
        return f"<WriteLogger(file={self._file!r})>"

    def msg(self, message: str) -> None:
        """
        Write and flush *message*.
        """
        with self._lock:
            until_not_interrupted(self._write, message + "\n")
            until_not_interrupted(self._flush)

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg


class WriteLoggerFactory:
    r"""
    Produce `WriteLogger`\ s.

    To be used with `structlog.configure`\ 's ``logger_factory``.

    :param file: File to print to. (default: `sys.stdout`)

    Positional arguments are silently ignored.

    .. versionadded:: 22.1
    """

    def __init__(self, file: TextIO | None = None):
        self._file = file

    def __call__(self, *args: Any) -> WriteLogger:
        return WriteLogger(self._file)


class BytesLogger:
    r"""
    Writes bytes into a file.

    :param file: File to print to. (default: `sys.stdout`\ ``.buffer``)

    Useful if you follow
    `current logging best practices <logging-best-practices>` together with
    a formatter that returns bytes (e.g. `orjson
    <https://github.com/ijl/orjson>`_).

    .. versionadded:: 20.2.0
    """
    __slots__ = ("_file", "_write", "_flush", "_lock")

    def __init__(self, file: BinaryIO | None = None):
        self._file = file or sys.stdout.buffer
        self._write = self._file.write
        self._flush = self._file.flush

        self._lock = _get_lock_for_file(self._file)

    def __getstate__(self) -> str:
        """
        Our __getattr__ magic makes this necessary.
        """
        if self._file is sys.stdout.buffer:
            return "stdout"

        if self._file is sys.stderr.buffer:
            return "stderr"

        raise PicklingError(
            "Only BytesLoggers to sys.stdout and sys.stderr can be pickled."
        )

    def __setstate__(self, state: Any) -> None:
        """
        Our __getattr__ magic makes this necessary.
        """
        if state == "stdout":
            self._file = sys.stdout.buffer
        else:
            self._file = sys.stderr.buffer

        self._write = self._file.write
        self._flush = self._file.flush
        self._lock = _get_lock_for_file(self._file)

    def __deepcopy__(self, memodict: dict[Any, Any] = {}) -> BytesLogger:
        """
        Create a new BytesLogger with the same attributes. Similar to pickling.
        """
        if self._file not in (sys.stdout.buffer, sys.stderr.buffer):
            raise copy.error(
                "Only BytesLoggers to sys.stdout and sys.stderr "
                "can be deepcopied."
            )

        newself = self.__class__(self._file)

        newself._write = newself._file.write
        newself._flush = newself._file.flush
        newself._lock = _get_lock_for_file(newself._file)

        return newself

    def __repr__(self) -> str:
        return f"<BytesLogger(file={self._file!r})>"

    def msg(self, message: bytes) -> None:
        """
        Write *message*.
        """
        with self._lock:
            until_not_interrupted(self._write, message + b"\n")
            until_not_interrupted(self._flush)

    log = debug = info = warn = warning = msg
    fatal = failure = err = error = critical = exception = msg


class BytesLoggerFactory:
    r"""
    Produce `BytesLogger`\ s.

    To be used with `structlog.configure`\ 's ``logger_factory``.

    :param file: File to print to. (default: `sys.stdout`\ ``.buffer``)

    Positional arguments are silently ignored.

    .. versionadded:: 20.2.0
    """
    __slots__ = ("_file",)

    def __init__(self, file: BinaryIO | None = None):
        self._file = file

    def __call__(self, *args: Any) -> BytesLogger:
        return BytesLogger(self._file)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import cProfile
import uuid
from contrast.utils.decorators import fail_loudly

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


MAX_ORIGINAL_PATH_LENTGH = 64


class Profiler(cProfile.Profile):
    def __init__(self, path):
        super().__init__()

        from contrast.agent.settings import Settings

        self.settings = Settings()
        self.path = path

    def __enter__(self):
        if self.settings.is_profiler_enabled:
            self.enable()
        return self

    def __exit__(self, *exc_info):
        if self.settings.is_profiler_enabled:
            self.disable()
            self._save_profile_data()

    @fail_loudly("Unable to save profile data")
    def _save_profile_data(self):
        safe_path = self._get_safe_path()
        filename = f"cprofile-{safe_path}-{uuid.uuid4().hex}.out"
        logger.debug("writing cprofile data to %s", filename)
        self.dump_stats(filename)

    def _get_safe_path(self):
        path = self.path[:MAX_ORIGINAL_PATH_LENTGH].strip("/").replace("/", "_") or "_"
        safe_path = "".join([c for c in path if c.isalpha() or c.isdigit() or c == "_"])
        return safe_path



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.config.base_config_rule import BaseConfigRule
from contrast.utils.decorators import fail_quietly


class BaseDjangoConfigRule(BaseConfigRule):
    # This should be overridden by child classes
    SETTINGS_VALUE = ""

    @property
    def name(self):
        raise NotImplementedError

    def get_snippet(self, value):
        """
        Build snippet to present in TS

        Eventually we could actually parse the settings file and provide context,
        but that seems like overkill right now.
        """
        if value is None:
            return f"[{self.SETTINGS_VALUE} not defined]"
        return f"{self.SETTINGS_VALUE} = {value!r}"

    def get_config_value(self, settings):
        return getattr(settings, self.SETTINGS_VALUE, None)

    def is_violated(self, value):
        raise NotImplementedError

    @fail_quietly("Failed to apply config rule")
    def apply(self, settings, config_module):
        value = self.get_config_value(settings)
        if not self.is_violated(value):
            return

        properties = self.create_properties(value, config_module.__file__)

        self.build_and_send_finding(
            properties, settings_module_path=config_module.__file__
        )

    def update_preflight_hash(self, hasher, settings_module_path=""):
        """
        Override method in base class for custom preflight hash generation
        """
        hasher.update(settings_module_path)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import importlib
import inspect
import sys
from types import FunctionType, ModuleType

from contrast.agent.policy import patch_manager
from contrast.agent.settings import Settings
from contrast.agent.policy.patch_manager import reverse_module_patches_by_name
from contrast.utils.decorators import fail_quietly
from contrast.utils.namespace import Namespace
from contrast.utils.patch_utils import repatch_imported_modules
from contrast_rewriter import ContrastRewriteLoader, populate_operator_module
from contrast_vendor import structlog as logging

# NOTE: it feels like overkill to store this in the policy registry right now,
# but we can always change this later if necessary.
REWRITE_MODULES = [
    "posixpath",
    "urllib.parse",
]

MODULE_REWRITE_SKIP_NAMES = [
    "__all__",
    "__name__",
    "__package__",
    "__module__",
    "__spec__",
    "__name__",
    "__builtins__",
    "__dict__",
    "__file__",
]

logger = logging.getLogger("contrast")


class policy_rewriter_state(Namespace):
    enabled: bool = False


def load_and_rewrite_module(module: ModuleType) -> ModuleType:
    """
    Returns a rewritten version of the given module

    (Note about frozen modules:)

    Frozen modules are modules whose bytecode is built into the interpreter
    itself for performance reasons. All of the modules required at interpreter
    startup are frozen in newer versions of Python (i.e. >= 3.10+).

    Frozen modules do not have access to the source code of module members such
    as functions. This means that calls toinspect.getsource(<frozen-module-name>.<member-name>)
    will fail. Since we need the source in order to perform rewrites, we need a
    non-frozen version of the module. We can achieve this by loading the module
    again and returning the new module object.
    """
    temp_name = "__contrast_temp." + module.__name__
    spec = importlib.util.spec_from_file_location(temp_name, module.__file__)

    module = importlib.util.module_from_spec(spec)
    ContrastRewriteLoader(temp_name, module.__file__).exec_module(module)

    assert temp_name not in sys.modules

    return module


@fail_quietly("Failed rewrite and patch function")
def rewrite_and_patch_function(
    module: ModuleType,
    name: str,
    function: FunctionType,
    new_module: ModuleType,
):
    # Some functions may already be patched by other policy, in which case we do not
    # want to rewrite them
    if patch_manager.is_patched(function):
        logger.debug("Skipping rewrite of already patched function: %s", name)
        return

    new_func = getattr(new_module, name)
    if new_func is None:
        logger.debug("No new function for %s. Skipping patch", name)
        return

    patch_manager.patch(module, name, new_func)


@fail_quietly("Failed to rewrite functions for module")
def rewrite_module_functions(module_name: str):
    module = sys.modules.get(module_name, None)
    if module is None:
        logger.debug(
            'Failed to rewrite functions in module "%s": module not loaded', module_name
        )
        return

    logger.debug("Applying rewriter policy to module: %s", module_name)

    populate_operator_module(module.__dict__)

    rewritten_module = load_and_rewrite_module(module)

    for name, member in inspect.getmembers(module):
        # If the unfrozen module doesn't have a function that is found in the
        # original module, it's probably the case that the function was added
        # by us (e.g. a function added by some other policy node).
        # (This check shouldn't really be necessary anymore)
        if not hasattr(rewritten_module, name):
            continue

        if name in MODULE_REWRITE_SKIP_NAMES:
            continue

        if patch_manager.is_patched(member):
            continue

        new_member = getattr(rewritten_module, name)
        if new_member is None:
            logger.debug("No new member for %s. Skipping patch", name)
            continue

        # This would apply to any member that was imported from another module
        if new_member is member:
            continue

        patch_manager.patch(module, name, new_member)


def apply_rewrite_policy(override_config: bool = False, rewrite_pathlib: bool = True):
    if policy_rewriter_state.enabled:
        logger.debug("Policy-based rewrites are already enabled")
        return

    if not (override_config or Settings().is_policy_rewriter_enabled):
        logger.debug("Skipping policy-based rewrites")
        return

    logger.debug("Applying policy-based rewrites")
    for module in REWRITE_MODULES:
        rewrite_module_functions(module)

    # Rewriting pathlib is problematic within the unit testing setting since
    # pytest relies heavily on pathlib and our patches wreak havoc. We enable
    # it by default but allow unit tests to disable it as necessary.
    if rewrite_pathlib and sys.version_info[:2] > (3, 10):
        rewrite_module_functions("pathlib")

    repatch_imported_modules()

    policy_rewriter_state.enabled = True


def reverse_rewrite_policy():
    for module in REWRITE_MODULES:
        reverse_module_patches_by_name(module)

    policy_rewriter_state.enabled = False



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast import AGENT_CURR_WORKING_DIR
from contrast.agent.protect.mixins.path_traversal_sink_features import (
    PathTraversalSinkFeatures,
)
from contrast.agent.protect.rule.base_rule import BaseRule, INFILTER
from collections.abc import Iterable

from contrast.utils.decorators import fail_quietly

PARENT_CHECK = ".."
SLASH = "/"
SAFE_PATHS = ["tmp", "public", "docs", "static", "template", "templates"]
WRITE_OPTIONS = ["w", "a"]


class PathTraversal(BaseRule, PathTraversalSinkFeatures):
    RULE_NAME = "path-traversal"

    def find_attack(self, candidate_string=None, analysis_stage=None, **kwargs):
        """
        Finds the attacker in the original string if present
        """
        attack = super().find_attack(
            candidate_string, analysis_stage=analysis_stage, **kwargs
        )
        if analysis_stage == INFILTER and candidate_string:
            attack = self.check_sink_features(candidate_string, attack)

        return attack

    def build_sample(self, evaluation, path, **kwargs):
        sample = self.build_base_sample(evaluation)
        if path is not None:
            sample.details["path"] = path
        return sample

    @fail_quietly(
        "Failed to run path traversal skip_protect_analysis", return_value=False
    )
    def skip_protect_analysis(self, user_input, args, kwargs):
        write = possible_write(args, kwargs)
        if write:
            # any write is a risk so we should not skip analysis
            return False

        return not actionable_path(user_input)

    def infilter_kwargs(self, user_input, patch_policy):
        return dict(method=patch_policy.method_name)


def possible_write(args, kwargs):
    return _possible_write_kwargs(kwargs) or _possible_write_args(args)


def _possible_write_kwargs(kwargs):
    mode = kwargs.get("mode", "")

    if not isinstance(mode, (str, bytes)):
        return False

    return mode and any([x in mode for x in WRITE_OPTIONS])


def _possible_write_args(args):
    if not isinstance(args, Iterable):
        return False

    return (
        len(args) > 1
        and args[1] is not None
        and isinstance(args[1], Iterable)
        and any([x in args[1] for x in WRITE_OPTIONS])
    )


def actionable_path(path):
    if not path or not isinstance(path, str):
        return False

    # moving up directory structure is a risk and hence actionable
    if path.find(PARENT_CHECK) > 1:
        return True

    if "/contrast/" in path or "/site-packages/" in path:
        return False

    if path.startswith(SLASH):
        for prefix in _safer_abs_paths():
            if path.startswith(prefix):
                return False
    else:
        for prefix in SAFE_PATHS:
            if path.startswith(prefix):
                return False

    return True


def _safer_abs_paths():
    return (
        [f"{AGENT_CURR_WORKING_DIR}/{item}" for item in SAFE_PATHS]
        if AGENT_CURR_WORKING_DIR
        else []
    )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from urllib.parse import parse_qs, urlencode, unquote

from contrast.agent.settings import Settings
from contrast.api.attack import ProtectResponse
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

MASK = "contrast-redacted-{}"
BODY_MASK = "contrast-redacted-body".encode("UTF-8")
SEMICOLON_URL_ENCODE_VAL = "%25"


class ActivityMasker:
    def __init__(self, ctx):
        self.ctx = ctx

        direct_ts_app_settings = getattr(Settings(), "direct_ts_app_settings", None)
        self.http_request = self.ctx.request
        if direct_ts_app_settings:
            self.mask_rules = direct_ts_app_settings.sensitive_data_masking_policy
        else:
            self.mask_rules = None

    def mask_sensitive_data(self):
        # Check if request is present and return if not
        if not self.http_request or not self.mask_rules:
            return

        logger.debug("Masker: masking sensitive data")

        self.mask_body()
        self.mask_query_string()
        self.mask_request_params()
        self.mask_request_cookies()
        self.mask_request_headers()

    def mask_body(self):
        # Check if mask_http_body is set to False or is None and skip if true
        if not self.mask_rules.get("mask_http_body"):
            return

        # Checks if body is not empty or null
        if self.http_request.body:
            self.http_request._masked_body = BODY_MASK

    def mask_query_string(self):
        query_string = self.http_request.query_string
        if query_string:
            self.http_request._masked_query_string = self.mask_raw_query(query_string)

    def mask_raw_query(self, query_string):
        qs_dict = parse_qs(query_string)
        self.mask_dictionary(qs_dict)
        return urlencode(qs_dict, doseq=True)

    def mask_request_params(self):
        params = self.http_request.params
        if not params:
            return

        self.http_request._masked_params = self.mask_dictionary(params)

    def mask_request_cookies(self):
        cookies = self.http_request.cookies
        if not cookies:
            return

        self.http_request._masked_cookies = self.mask_dictionary(cookies)

    def mask_request_headers(self):
        headers = self.http_request.headers
        if not headers:
            return

        self.http_request._masked_headers = self.mask_dictionary(headers)

    def mask_dictionary(self, d):
        if not d:
            return None

        d_copy = dict(d).copy()

        for k, v in d_copy.items():
            if self.find_value_index_in_rules(k.lower()) == -1:
                continue

            if isinstance(v, list):
                self.mask_values(k, v, d_copy, self.ctx.attacks)
            else:
                self.mask_hash(k, v, d_copy, self.ctx.attacks)
        return d_copy

    def mask_values(self, k, v, d, attacks):
        for idx, item in enumerate(v):
            if self.mask_rules.get("mask_attack_vector") and self.is_value_vector(
                attacks, item
            ):
                d[k][idx] = MASK.format(k.lower())
            if not self.is_value_vector(attacks, item):
                d[k][idx] = MASK.format(k.lower())

    def mask_hash(self, k, v, d, attacks):
        if self.mask_rules.get("mask_attack_vector") and self.is_value_vector(
            attacks, v
        ):
            d[k] = MASK.format(k.lower())
        if not self.is_value_vector(attacks, v):
            d[k] = MASK.format(k.lower())

    def mask_pair(self, k, v, d, attacks):
        for idx, item in enumerate(v):
            if self.mask_rules.get("mask_attack_vector") and self.is_value_vector(
                attacks, item
            ):
                d[k].values[idx] = MASK.format(k.lower())
            if not self.is_value_vector(attacks, item):
                d[k].values[idx] = MASK.format(k.lower())

    def is_value_vector(self, attacks, value):
        if not attacks or not value:
            return False

        for attack in attacks:
            if self.is_value_in_sample(attack.samples, value):
                return attack.response != ProtectResponse.NO_ACTION

        return False

    def is_value_in_sample(self, samples, value):
        if not samples:
            return False

        # Setting this to remove url encoding of header and cookie values
        value = unquote(value)

        for sample in samples:
            if sample.user_input.value == value:
                return True
        return False

    def find_value_index_in_rules(self, s):
        index = -1
        # When looking for header it replaces '_' with '-' and I don't want to risk not
        # properly matching to the rules
        s = s.replace("-", "_")
        for rule in self.mask_rules.get("rules"):
            try:
                index = rule.get("keywords").index(s)
                break
            except ValueError:
                index = -1

        return index



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Using logger.info/debug/someOtherLevel() is not supported in this module. In order to get the correct
frame info, we must skip over functions called in this module and in vendored structlog. If logging is attempted,
incorrect frame info will be displayed on the log message if used in this file.

Use print(...) instead
"""
import asyncio
import sys
from os import getpid
from os.path import basename
import threading

import contrast
from contrast_vendor.structlog._frames import _find_first_app_frame_and_name
from contrast_vendor import structlog
from contrast.utils.configuration_utils import get_hostname

LOGGING_TO_BUNYAN_LOG_LEVEL_CONVERSION = {
    "critical": 60,
    "error": 50,
    "warning": 40,
    "info": 30,
    "debug": 20,
}


def add_hostname(logger, method_name, event_dict):
    event_dict["hostname"] = get_hostname()

    return event_dict


def add_pid(logger, method_name, event_dict):
    event_dict["pid"] = getpid()

    return event_dict


def add_thread_id(logger, method_name, event_dict):
    event_dict["thread_id"] = threading.get_ident()

    return event_dict


def add_request_id(logger, method_name, event_dict):
    context = contrast.CS__CONTEXT_TRACKER.current()
    obj_id = -1

    if context:
        obj_id = id(context)

    event_dict["request_id"] = obj_id

    return event_dict


def rename_key(old_name, new_name):
    def event_key_to_msg(logger, method_name, event_dict):
        """
        msg is a required key for bunyan parsing. The event key is renamed to msg
        """

        value = event_dict.get(old_name)
        if value and not event_dict.get(new_name):
            event_dict[new_name] = value
            del event_dict[old_name]

        return event_dict

    return event_key_to_msg


def add_bunyan_log_level(logger, log_level, event_dict):
    """
    This Processor must be installed AFTER structlog.stdlib.add_log_level.
    structlog.stdlib.add_log_level adds level: "info/debug/...". This function
    converts that string to the bunyan integer value equivalent (whenever possible).
    """
    if log_level == "warn":
        # The stdlib has an alias
        log_level = "warning"

    new_value = LOGGING_TO_BUNYAN_LOG_LEVEL_CONVERSION.get(log_level, None)

    if new_value:
        event_dict["level"] = new_value

    return event_dict


def add_v(logger, method_name, event_dict):
    """
    Required key for bunyan log parsing
    """
    event_dict["v"] = 0

    return event_dict


def add_frame_info(logger, method_name, event_dict):
    """
    Adds filename, function name and line number based on where the logger is called
    """
    ignore_frames = [
        "contrast_vendor.structlog",
        "contrast.utils.loggers.structlog",
        "logging",
    ]

    frame_info = _find_first_app_frame_and_name(ignore_frames)

    if frame_info and frame_info[0]:
        frame = frame_info[0]

        event_dict[
            "frame_info"
        ] = f"{basename(frame.f_code.co_filename)}:{frame.f_code.co_name}:{frame.f_lineno}"

    return event_dict


def add_progname(logger, method_name, event_dict):
    """
    progname is the name of the process the agents uses in logs.
    The default value is Contrast Agent. progname will be used
    as the name of the logger as seen in the logs.
    """
    field = "name"
    current_handler = logger.handlers[0]

    if hasattr(current_handler.filters[0], field):
        progname = current_handler.filters[0].progname

        if progname:
            event_dict[field] = progname

    return event_dict


def add_asyncio_info(logger, method_name, event_dict):
    try:
        current_task = asyncio.current_task()

        if sys.version_info[:2] > (3, 7):
            # get_name and get_coro are only supported on 3.8+

            # If no name has been explicitly assigned to the Task, the default asyncio Task implementation
            # generates a default name during instantiation.
            event_dict["asyncio_task_name"] = current_task.get_name()

            current_coro = current_task.get_coro()
            if hasattr(current_coro, "__name__"):
                event_dict["asyncio_coro_name"] = current_coro.__name__

        event_dict["asyncio_task_id"] = id(current_task)
    except Exception:
        # This can happen when there is no running event loop
        pass

    return event_dict


def init_structlog():
    """
    Configures structlog -- must be called AFTER logging module is configured
    """
    structlog.configure(
        # Each processor is called from the top down and can modify the event_dict passed to it
        processors=[
            add_bunyan_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt="iso"),
            # rename_key must be called after timestamp is added by TimeStamper
            rename_key("timestamp", "time"),
            rename_key("event", "msg"),
            add_v,
            add_hostname,
            add_pid,
            add_thread_id,
            add_request_id,
            add_frame_info,
            add_progname,
            add_asyncio_info,
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.JSONRenderer(),
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
ALL_ARGS = "ALL_ARGS"
ALL_KWARGS = "ALL_KWARGS"

OBJECT = "OBJ"
OBJECT_KEY = OBJECT[0]

RETURN = "RETURN"
RETURN_KEY = RETURN[0]

ARG = "ARG"
KWARG = "KWARG"
ARG_OR_KWARG = "P"

ALL_TYPE = "A"
CREATION_TYPE = "CREATION"
TRIGGER_TYPE = "TRIGGER"
TO_MARKER = "2"

SCOPE = "contrast"

CONTRAST_KEY = "contrast"
SOURCES_KEY = "sources"
PROPAGATION_KEY = "propagators"
TRIGGERS_KEY = "triggers"
DEADZONE_KEY = "deadzones"
REWRITES_KEY = "rewrites"

DB_SOURCE_TYPE = "TAINTED_DATABASE"

DUNDER = "__"

CURRENT_FINDING_VERSION = 4

# For rare case of dataflow rules without routes
MINIMUM_FINDING_VERSION = 3

# For PolicyNode
VALID_TAGS = [
    "XML_ENCODED",
    "XML_DECODED",
    "HTML_ENCODED",
    "HTML_DECODED",
    "URL_ENCODED",
    "URL_DECODED",
    "CSS_ENCODED",
    "CSS_DECODED",
    "BASE64_ENCODED",
    "BASE64_DECODED",
    "JAVASCRIPT_ENCODED",
    "JAVASCRIPT_DECODED",
    "JAVA_ENCODED",
    "JAVA_DECODED",
    "CSV_ENCODED",
    "CSV_DECODED",
    "SQL_ENCODED",
    "SQL_DECODED",
    "LDAP_ENCODED",
    "LDAP_DECODED",
    "XPATH_ENCODED",
    "XPATH_DECODED",
    "OS_ENCODED",
    "OS_DECODED",
    "VBSCRIPT_ENCODED",
    "VBSCRIPT_DECODED",
    "POTENTIAL_SANITIZED",
    "POTENTIAL_VALIDATED",
    "XMLIF_VALIDATED_XXE",
    "NO_CONTROL_CHARS",
    "JS_ENCODED",
    "XSS_ENCODED",
    "LIMITED_CHARS",
    "SAFE_PATH",
    "CUSTOM",
    "CUSTOM_ENCODED",
    "CUSTOM_ENCODED_CMD_INJECTION",
    "CUSTOM_ENCODED_EXPRESSION_LANGUAGE_INJECTION",
    "CUSTOM_ENCODED_HEADER_INJECTION",
    "CUSTOM_ENCODED_HQL_INJECTION",
    "CUSTOM_ENCODED_LDAP_INJECTION",
    "CUSTOM_ENCODED_LOG_INJECTION",
    "CUSTOM_ENCODED_NOSQL_INJECTION",
    "CUSTOM_ENCODED_PATH_TRAVERSAL",
    "CUSTOM_ENCODED_REDOS",
    "CUSTOM_ENCODED_REFLECTED_XSS",
    "CUSTOM_ENCODED_REFLECTION_INJECTION",
    "CUSTOM_ENCODED_SMTP_INJECTION",
    "CUSTOM_ENCODED_SQL_INJECTION",
    "CUSTOM_ENCODED_SSRF",
    "CUSTOM_ENCODED_STORED_XSS",
    "CUSTOM_ENCODED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_ENCODED_UNSAFE_CODE_EXECUTION",
    "CUSTOM_ENCODED_UNSAFE_READLINE",
    "CUSTOM_ENCODED_UNSAFE_XML_DECODE",
    "CUSTOM_ENCODED_UNTRUSTED_DESERIALIZATION",
    "CUSTOM_ENCODED_UNVALIDATED_FORWARD",
    "CUSTOM_ENCODED_UNVALIDATED_REDIRECT",
    "CUSTOM_ENCODED_XPATH_INJECTION",
    "CUSTOM_ENCODED_XXE",
    "CUSTOM_SECURITY_CONTROL_APPLIED",
    "CUSTOM_VALIDATED",
    "CUSTOM_VALIDATED_CMD_INJECTION",
    "CUSTOM_VALIDATED_EXPRESSION_LANGUAGE_INJECTION",
    "CUSTOM_VALIDATED_HEADER_INJECTION",
    "CUSTOM_VALIDATED_HQL_INJECTION",
    "CUSTOM_VALIDATED_LDAP_INJECTION",
    "CUSTOM_VALIDATED_LOG_INJECTION",
    "CUSTOM_VALIDATED_NOSQL_INJECTION",
    "CUSTOM_VALIDATED_PATH_TRAVERSAL",
    "CUSTOM_VALIDATED_REDOS",
    "CUSTOM_VALIDATED_REFLECTED_XSS",
    "CUSTOM_VALIDATED_REFLECTION_INJECTION",
    "CUSTOM_VALIDATED_SMTP_INJECTION",
    "CUSTOM_VALIDATED_SQL_INJECTION",
    "CUSTOM_VALIDATED_SSRF",
    "CUSTOM_VALIDATED_STORED_XSS",
    "CUSTOM_VALIDATED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_VALIDATED_UNSAFE_CODE_EXECUTION",
    "CUSTOM_VALIDATED_UNSAFE_READLINE",
    "CUSTOM_VALIDATED_UNSAFE_XML_DECODE",
    "CUSTOM_VALIDATED_UNTRUSTED_DESERIALIZATION",
    "CUSTOM_VALIDATED_UNVALIDATED_FORWARD",
    "CUSTOM_VALIDATED_UNVALIDATED_REDIRECT",
    "CUSTOM_VALIDATED_XPATH_INJECTION",
    "CUSTOM_VALIDATED_XXE",
    "DATABASE_WRITE",
]

NO_NEWLINES = "NO_NEWLINES"
CROSS_SITE = "CROSS_SITE"

VALID_SOURCE_TAGS = [NO_NEWLINES, "UNTRUSTED", CROSS_SITE, "LIMITED_CHARS"]

JSON_NAME = "name"
JSON_MODULE = "module"
JSON_CLASS_NAME = "class_name"
JSON_INSTANCE_METHOD = "instance_method"
JSON_METHOD_NAME = "method_name"
JSON_SOURCE = "source"
JSON_TARGET = "target"
JSON_TAGS = "tags"
JSON_POLICY_PATCH = "policy_patch"
JSON_GOOD_VALUE = "good_value"
JSON_BAD_VALUE = "bad_value"
JSON_TYPE = "type"
JSON_SOURCE_NAME = "source_name"
JSON_ACTION = "action"
JSON_UNTAGS = "untags"
JSON_PACKAGE = "package"
JSON_VERSION = "version"
JSON_BUILTIN = "builtin"
JSON_PROTECT_MODE = "protect_mode"
JSON_UNSAFE_DEFAULT = "unsafe_default"
JSON_CONFIG_OPTION = "config_option"

# End of PolicyNode

TRIGGER_ACTION_DEFAULT = "DEFAULT"
TRIGGER_ACTION_REDOS = "REDOS"
TRIGGER_ACTION_SSRF = "SSRF"
TRIGGER_ACTION_FROMSTRING = "FROMSTRING"
TRIGGER_ACTION_SUBPROCESS = "SUBPROCESS"
TRIGGER_ACTION_OPENAI = "OPENAI"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

from contrast.utils import get_installed_distributions
from contrast.utils.singleton import Singleton
from contrast.utils.library_reader.utils import (
    get_top_level_directories,
    get_data,
    get_hash,
    SITE_PACKAGES_DIR,
    DIST_PACKAGES_DIR,
    get_top_level_directories_namespace_pkg,
)

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class DistributionContext(Singleton):
    """
    This class is used to maintain state between calls to python import functions.
    The main reason why we need to do all this is because Python doesn't provide
    us with a function that maps a loaded file name
    (e.g /Users/nliccione/lib/python3.7/site-packages/django/https.py)
    to the distribution it belongs to.
    """

    class TopLevelDirNode:
        def __init__(self, top_level_dir, dist):
            version, _, _ = get_data(dist)
            name = str(dist).split(" ")[0]

            self.namespace_tlds = set()
            self.dist_hash = get_hash(name, version)
            self.dist = dist
            self.distribution_install_dir = dist.location
            self.full_package_path = os.path.join(dist.location, top_level_dir)

    def init(self):
        # dictionary containing mapping of top level directory name -> ( TopLevelDirNode, )
        self.top_level_dir_map = {}
        self.installed_dists = get_installed_distributions()

        for dist in self.installed_dists:
            self._build_tld_entries_from_dist(dist)

    def _get_dist_ctx_from_file_path(self, file_path):
        """
        This function returns the TopLevelDirNode that represents the given file_path.
        @param file_path: Path to the file (e.g /Users/nliccione/lib/python3.7/site-packages/django/https.py)
        @rtype: TopLevelDirNode
        """
        ret = None

        if not file_path:
            return ret

        # Using this as an example: /Users/nliccione/lib/python3.7/site-packages/django/https.py
        # package_dirs[1] should be "django/https.py" where django is the top level importable package
        # and https.py is the specific module to be loaded (top level module, then https.py after)
        package_dirs = file_path.rsplit(SITE_PACKAGES_DIR)

        if len(package_dirs) < 2:
            package_dirs = file_path.rsplit(DIST_PACKAGES_DIR)
            if len(package_dirs) < 2:
                return None

        # Get the top level importable module/package from the file path and set it to top_level_name
        # Using the django example top_level_name should be "django"
        top_level_relative_dir = package_dirs[1]
        top_level_dirs = top_level_relative_dir.split(os.sep)
        top_level_name = top_level_dirs[0]

        if top_level_name:
            top_level_node = None

            # If the top level name is a module file than we report on that because that is the top level import
            if top_level_name.endswith(".py") or top_level_name.endswith(".pyc"):
                top_level_name = os.path.splitext(top_level_name)[0]

            nodes = self.top_level_dir_map.get(top_level_name, [None])
            if len(nodes) == 1:
                top_level_node = nodes[0]
            elif len(top_level_dirs) > 1:
                # Handling a namespace package where multiple distributions share the same top level directory
                # tests/samples/sample_packages/sample_namespace_pkg1 and sample_namespace_pkg2
                # are examples of this case
                namespace_tld = top_level_dirs[1]
                for node in nodes:
                    if namespace_tld in node.namespace_tlds:
                        top_level_node = node
                        break

            # Once we get the TopLevelDirNode associated with the top_level_name found in file_path, we then verify
            # TopLevelDirNode is the node that actually matches file_path based on the path to the distribution
            # found in TopLevelDirNode
            if top_level_node and file_path.startswith(
                top_level_node.full_package_path
            ):
                ret = top_level_node

        return ret

    def get_dist_hash_from_file_path(self, file_path):
        """
        Convenience wrapper around _get_dist_ctx_from_file_path for getting the distribution hash for a file
        @param file_path: Path to the file (e.g /Users/nliccione/lib/python3.7/site-packages/django/https.py)
        @return dist_hash of the distribution associated with file_path
        @rtype: string
        """
        dist_hash = None

        ctx = self._get_dist_ctx_from_file_path(file_path)
        if ctx:
            dist_hash = ctx.dist_hash

        return dist_hash

    @staticmethod
    def _add_namespace_tlds_to_node(tld, node):
        dist = node.dist

        tld_namespace_dirs = get_top_level_directories_namespace_pkg(dist, tld)

        if len(tld_namespace_dirs) == 0:
            logger.debug(
                "WARNING - namespace package does not have a RECORD file. Unable to parse \
                top level directory for the distribution: %s",
                str(dist),
            )
        else:
            node.namespace_tlds = tld_namespace_dirs

    def _build_tld_entries_from_dist(self, dist):
        """
        This function builds the data structure containing the mapping of a top level package/module name
        belonging to a specific distribution.
        For example, if /Users/nliccione/lib/python3.7/site-packages/django/https.py is the file that is imported, an entry in self.top_level_dir_map
        will look like this: self.top_level_dir_map = {"django": [ TopLevelDirNode, ]}. When https.py is imported, we check to see if "django"
        exists in self.top_level_dir_map. If it does, and the file path found in sys.module matches TopLevelDirNode.full_package_path, than
        we know the imported file belongs to the distribution cached in self.top_level_dir_map.

        This function also handles namespace distributions. Namespace distributions share a top level import name
        (e.g zope namespace where zope.deprecation is a distribution using that namespace). As a result, we need to have an additional
        data structure to maintain the list of top level directories or file names belonging to that namespace.

        @param dist: Distribution object used to determine how many top level importable packages/modules exist in it
        @type dist: pkg_resources.DistInfoDistribution
        """
        for tld in get_top_level_directories(dist):
            # The namespace dist we support is building a namespace package
            # using the pkg_resources method of defining one
            # https://packaging.python.org/guides/packaging-namespace-packages/#pkg-resources-style-namespace-packages
            new_node = DistributionContext.TopLevelDirNode(tld, dist)

            namespace_dists = self.top_level_dir_map.get(tld, None)
            if namespace_dists:
                self._add_namespace_tlds_to_node(tld, new_node)

                if len(namespace_dists) == 1:
                    # Handle case where we didn't know we have a namespace dist before.
                    # Now that we know we need to go back and add the tlds for that namespace node
                    prev_node = namespace_dists[0]
                    self._add_namespace_tlds_to_node(tld, prev_node)

                namespace_dists.append(new_node)
            else:
                self.top_level_dir_map[tld] = [new_node]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.triggers.secure_flag_missing_rule import (
    SecureFlagMissingRule,
)
from contrast.agent.policy.registry import register_trigger_rule


secure_flag_missing_triggers = [
    {
        "module": "pyramid.session",
        "method_name": "BaseCookieSessionFactory",
        "source": "KWARG:secure",
        "unsafe_default": True,
        "policy_patch": False,
    },
    {
        # WSGI pluggable for sessions
        "module": "beaker.session",
        "class_name": "Session",
        "method_name": "__init__",
        "source": "KWARG:secure",
        "unsafe_default": True,
    },
    {
        # Set custom response cookie and set its secure attribute
        "module": "falcon",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:secure",
        "unsafe_default": True,
    },
    {
        # Used by bottle for sessions
        "module": "bottle_session",
        "class_name": "Session",
        "method_name": "__init__",
        "source": "ARG_4,KWARG:cookie_secure",
        "unsafe_default": True,
    },
    {
        # Used by fastapi for cookies
        "module": "starlette.responses",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "ARG_7,KWARG:secure",
        "unsafe_default": True,
    },
    {
        # Set custom response cookie and set its secure attribute
        "module": "aiohttp.web",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:secure",
        "unsafe_default": True,
    },
]


register_trigger_rule(
    SecureFlagMissingRule.from_nodes(
        "secure-flag-missing",
        secure_flag_missing_triggers,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.assess.duck_utils import len_or_zero


class Preshift(object):
    """
    Holder class for information prior to shifting
    """

    def __init__(self, self_obj, call_args, call_kwargs):
        self.obj = self_obj
        self.obj_length = len_or_zero(self_obj)
        self.args = call_args or []
        self.kwargs = call_kwargs or {}

    def __repr__(self):
        return f"<Preshift({self.obj}, {self.args}, {self.kwargs})>"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent import scope

from contrast.agent.request import Request
from contrast.agent.settings import Settings
from contrast.api.route_coverage import Route
from contrast_vendor import structlog as logging
from contrast.utils.digest_utils import Digest
from contrast.utils.decorators import cached_property, fail_quietly

logger = logging.getLogger("contrast")


class RequestContext(object):
    def __init__(self, environ):
        scope.enter_contrast_scope()

        self.request = Request(environ)
        self.response = None

        # This contains any Observed Library Usage seen during the life of this request.
        self.observed_library_usage = None

        # This contains input exclusions that should always apply regardless of rule on this request
        self.input_exclusions = None

        # This contains exclusions that only apply at trigger time based on the source event
        self.input_exclusions_trigger_time = None

        self.excluded_assess_rules = []
        self.excluded_protect_rules = []

        # For protect: store attacks made during a request to report at the end
        self.attacks = []

        # For assess: store findings made during a request to report at the end
        self.findings = []

        self.user_input_analysis = []
        self.response = None
        self.do_not_track = False

        self.view_func = None
        self.view_func_str = None
        # to be populated with a Route instance
        self.current_route = None

        self.observed_route = Route("", "", "")
        self.database_info = []
        self.source_count = 0
        self.propagation_count = 0

        self.max_sources_logged = False
        self.max_propagators_logged = False

        scope.exit_contrast_scope()

    @cached_property
    @fail_quietly("Unable to compute request hash")
    def hash(self):
        """
        Generates the hash checksum for the request. Converts the request method, uri,
        param names and content length to CRC checksum and returns string representation

        :return: str
        """
        hasher = Digest()

        hasher.update(self.request.method)
        hasher.update(self.request.get_normalized_uri())
        for key in self.request.params:
            hasher.update(key)

        hasher.update(self.request.content_length)
        return hasher.finish()

    def stop_source_creation(self, context, source_type, source_name):
        """
        Compare `source_count` to `max_context_source_events` config option
        :return: true if source_count within this request is equal to or
                 greater than configured threshold for source creation
        """
        if self.max_sources_logged:
            return True

        if context.input_exclusions and Settings().exclusions.evaluate_input_exclusions(
            context, source_type, source_name
        ):
            return True

        threshold_reached = self.source_count >= Settings().max_sources
        if threshold_reached:
            logger.warning(
                "Will not create more sources in this request. %s sources reached",
                self.source_count,
            )
            self.max_sources_logged = True

        return threshold_reached

    @property
    def stop_propagation(self):
        """
        Compare `source_count` to `max_propagation_events` config option
        :return: true if propagation_count within this request is equal to or
                 greater than configured threshold for propagation
        """
        if self.max_propagators_logged:
            return True

        threshold_reached = self.propagation_count >= Settings().max_propagation
        if threshold_reached:
            logger.warning(
                "Will not propagate any more in this request. %s propagations reached.",
                self.propagation_count,
            )
            self.max_propagators_logged = True

        return threshold_reached

    @property
    def propagate_assess(self):
        # TODO: PYT-644 move this property of out this class?
        return Settings().is_assess_enabled() and not scope.in_scope()

    def source_created(self):
        """
        Increase the running count of sources created in this request
        """
        self.source_count += 1

    def propagated(self):
        """
        Increase the running count of propagations created in this request
        """
        self.propagation_count += 1

    def extract_response_to_context(self, response):
        """
        Append response to request context if required for analysis.

        Protect always saves the Response as it is needed for PostFilter; however we may be able
        to optimize this by checking if any rules that act on teh response require postfilter.

        Assess optimizes this by only saving the Response if response scanning, which is a form
        of selective sampling, is enabled for this rule.

        :response: Subclass of BaseResponseWrapper
        """
        if (
            not Settings().is_protect_enabled()
            and not Settings().response_scanning_enabled
        ):
            return

        self.response = response



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
from collections import namedtuple
import os

from contrast import AGENT_CURR_WORKING_DIR
import contrast.utils.configuration_utils as utils

from contrast.utils.decorators import cached_property
from contrast.utils.loggers import (
    DEFAULT_PROGNAME,
    DEFAULT_SECURITY_LOG_PATH,
    DEFAULT_SECURITY_LOG_LEVEL,
)
from contrast_vendor import structlog as logging
from contrast.utils.singleton import Singleton

logger = logging.getLogger("contrast")

SPLIT_CWD = AGENT_CURR_WORKING_DIR.split(os.sep)
APP_NAME = SPLIT_CWD[len(SPLIT_CWD) - 1]
ENV_PREFIX = "CONTRAST"


DEFAULT_TEAMSERVER_TYPE = "EOP"
PUBLIC_SASS_TEAMSERVER_TYPES = {
    "app.contrastsecurity.com": "SAAS_DEFAULT",
    "eval.contrastsecurity.com": "SAAS_POV",
    "ce.contrastsecurity.com": "SAAS_CE",
    "app.contrastsecurity.jp": "SAAS_TOKYO",
    "cs": "SAAS_",
}

PRIVATE_SASS_TEAMSERVER_TYPE = "SAAS_CUSTOM"
PRIVATE_SASS_DOMAIN = ["contrastsecurity.com", "contrastsecurity.jp"]

TESTING_TEAMSERVER_TYPE = "TESTING"
TESTING_TEAMSERVER_NAMES = [
    "darpa",
    "alpha.contrastsecurity.com",
    "apptwo.contrastsecurity.com",
    "teamserver-staging.contsec.jp",
    "teamserver-staging.contsec.com",
    "security-research.contrastsecurity.com",
    "teamserver-ops.contsec.com",
]


ConfigOption = namedtuple(
    "ConfigOption", ["canonical_name", "default_value", "type_cast"]
)
ConfigOptionSrc = namedtuple("ConfigOptionSrc", ["name", "source"])

ENVIRONMENT_VARIABLE_SRC = "ENVIRONMENT_VARIABLE"
USER_CONFIGURATION_FILE_SRC = "USER_CONFIGURATION_FILE"
DEFAULT_VALUE_SRC = "DEFAULT_VALUE"
CONTRAST_UI_SRC = "CONTRAST_UI"
ASSESS_DISABLED_RULE_CONFIG_KEY = "assess.rules.disabled_rules"


class AgentConfig(object):
    def __init__(self):
        self.yaml_filename = None
        self.config_status = None
        self._config_source = {}
        self._config = utils.load_yaml_config()

        if not self._config:
            logger.info("No YAML config found; using default settings")
            self._config = {}

        self.yaml_filename = self._config.get("loaded_configuration_filename", None)
        if self.yaml_filename is not None:
            del self._config["loaded_configuration_filename"]

        self._config = utils.flatten_config(self._config)

        self.override_configs()

        self.check_for_api_config()

        # This stores the session ID sent from TS. We have to separate this value from the value set in the yaml config
        # because the value sent from TS always takes precedence. This value is only sent from TS
        # in response to ApplicationStartup
        self._ts_session_id = None

    def get_src(self, key, default_value=None):
        return self._config_source.get(key, default_value)

    def get(self, key, default_value=None):
        return self._config.get(key, default_value)

    def put(self, key, value, src=None):
        if not key or key not in self._config:
            return

        if src is not None:
            self._config_source[key] = src

        self._config[key] = value

    def get_redacted_config(self):
        """
        The purpose of this function is to return a copy of the config with redacted values
        """
        sensitive_keys = [
            "api.service_key",
            "api.user_name",
            "api.api_key",
            "api.proxy.url",
        ]

        redacted_config = self._config.copy()
        for key in sensitive_keys:
            if redacted_config[key]:
                redacted_config[key] = "**REDACTED**"

        return redacted_config

    def log_config(self):
        """
        Log configuration values.

        Config values deemed sensitive are logged as **REDACTED** unless unassigned.
        """
        redacted_config = self.get_redacted_config()

        logger.info("Current Configuration", **redacted_config)

    def _validate_agent_config(self):
        """
        At this point in configuration initialization every item in self._config should have an
        associated source in self._config_source (e.g set by DEFAULT VALUE or ENVIRONMENT VARIABLE)

        If it does not have a source than the assumption is made that the key does not have an entry in
        any of the default_options set in each of the top level configs.

        If that is the case this is either: A new configuration option that should be
        added to default_options OR a misconfiguration by the user
        """
        for k in self._config.keys():
            if self._config_source.get(k, None) is None:
                if self.config_status is None:
                    self.config_status = f"Invalid {self._config[k]} on {k}"
                    return

    def override_configs(self):
        """
        For each class representing different parts of the config (Agent, etc),
        take the current config and apply any overriding logic such as overriding with
        environment keys/values.

        At this time the order of precedence is:
            os.env > Config yaml

        Meaning that a config value defined in os.environ must be used instead of
        the same config defined in contrast_security.yaml

        Note that CLI args (sys.argv) are not supported. This may change if the
        agent becomes a runner.
        """
        builders = [Api, Agent, Application, Assess, Inventory, Protect, Root, Server]

        for builder in builders:
            status = builder().build(self._config, self._config_source)
            if status:
                # An error here indicates a failure with the type cast of the value
                self.config_status = status
                return

        self._validate_agent_config()

    def check_for_api_config(self):
        """
        Validate api configurations were set.

        If any api configs are missing, log at ERROR and disable the agent.

        Returns the missing values or an empty list if none were missing.
        """

        api_keys = [
            "api.url",
            "api.service_key",
            "api.api_key",
            "api.user_name",
        ]

        missing_values = []
        for key in api_keys:
            val = self.get(key)
            if not val:
                missing_values.append(key)

        if missing_values:
            msg = (
                f"Missing a required connection value for: {', '.join(missing_values)}"
            )
            logger.error(msg)
            self.put("enable", False)

        return missing_values

    @property
    def session_id(self):
        return (
            self._ts_session_id
            if self._ts_session_id is not None
            else self.get("application.session_id", "")
        )

    @session_id.setter
    def session_id(self, session_id):
        self._ts_session_id = session_id
        self._config_source["application.session_id"] = ConfigOptionSrc(
            name="contrast.application.session_id", source=CONTRAST_UI_SRC
        )
        logger.debug("Set session_id to %s", session_id, direct_to_teamserver=1)

    def get_session_metadata(self):
        return self.get("application.session_metadata", "")

    @property
    def app_code(self):
        return self.get("application.code", "")

    @property
    def app_metadata(self):
        return self.get("application.metadata", "")

    @property
    def app_group(self):
        return self.get("application.group", "")

    @property
    def app_tags(self):
        return self.get("application.tags", "")

    @property
    def assess_tags(self):
        return self.get("assess.tags", "")

    @cached_property
    def is_request_audit_enabled(self):
        return self.get("api.request_audit.enable")

    @cached_property
    def assess_enabled(self):
        return self.get("assess.enable")

    @cached_property
    def should_rewrite(self):
        return self.assess_enabled and self.get("agent.python.rewrite")

    @cached_property
    def should_apply_policy_rewrites(self):
        return self.should_rewrite and self.get("agent.python.enable_policy_rewrites")

    @cached_property
    def enable_automatic_middleware(self) -> bool:
        return self.get("agent.python.enable_automatic_middleware", True)

    @cached_property
    def application_activity_polling_interval(self):
        return self.get("agent.polling.app_activity_ms", 30_000)

    @cached_property
    def server_settings_poll_interval(self):
        return self.get("agent.polling.server_settings_ms", 30_000)

    @cached_property
    def teamserver_type(self):
        url = self.get("api.url")

        if url is None:
            return ""

        for name, ts_type in PUBLIC_SASS_TEAMSERVER_TYPES.items():
            if name in url:
                # This regex matches https://cs001.contrastsecurity.com
                r_url = re.match(r".*(cs\d{3})\..*", url)
                if r_url:
                    # and gets the second group which is the matched in brackets cs001
                    # adds it to the SAAS_ type and becomes SAAS_CS001
                    numbers = r_url.group(1).upper()

                    return ts_type + numbers

                return ts_type

        for name in TESTING_TEAMSERVER_NAMES:
            if name in url:
                return TESTING_TEAMSERVER_TYPE

        for name in PRIVATE_SASS_DOMAIN:
            if name in url:
                return PRIVATE_SASS_TEAMSERVER_TYPE

        return DEFAULT_TEAMSERVER_TYPE


class ConfigBuilder(object):
    TOP_LEVEL = ""

    def __init__(self):
        self.default_options = []

    def build(self, config, config_src):
        """
        Given a dict config, iterate over the default_options and check if
        the corresponding config key/value should be either :
        1. replaced by an existing env var
        2. keep existing config key/val but type-cast the value
        3. add a new key/default_value to the config

        :param config: dict config
        :param config_src: dict where each config option is a key that maps to a ConfigOptionSrc
        :return: str if error was set, config dict is updated pass by reference
        """
        sampling_lock_config = None
        if isinstance(self, Assess):
            sampling_lock_config = SamplingLockConfig()

        for option_name, default_val, type_cast in self.default_options:
            dot_alt = self._dot_alternative(option_name)
            underscore_alt = self._underscore_alternative(option_name)

            env_override = utils.get_env_key(underscore_alt)
            if env_override:
                try:
                    # replace the config value with the os.env value and apply type-cast
                    env_override = type_cast(env_override)
                except Exception as e:
                    logger.exception("Failed to initialize config")
                    return f"Invalid value on {dot_alt} - {e}"

                config[dot_alt] = env_override
                config_src[dot_alt] = ConfigOptionSrc(
                    name=underscore_alt, source=ENVIRONMENT_VARIABLE_SRC
                )
                continue

            if "sampling" in option_name and config.get(dot_alt):
                sampling_lock_config.set_sampling_config(option_name.split(".")[-1])

            if dot_alt in config:
                # update the config value with a type-cast value
                val = config[dot_alt]

                try:
                    val = type_cast(val)
                except Exception as e:
                    logger.exception("Failed to initialize config")
                    return f"Invalid value on {dot_alt} - {e}"

                config[dot_alt] = val
                config_src[dot_alt] = ConfigOptionSrc(
                    name=dot_alt, source=USER_CONFIGURATION_FILE_SRC
                )
                continue

            # add a new key/default_value to config
            config[dot_alt] = default_val
            config_src[dot_alt] = ConfigOptionSrc(
                name=dot_alt, source=DEFAULT_VALUE_SRC
            )

        return None

    def _underscore_alternative(self, key):
        return "__".join(
            [x for x in [ENV_PREFIX, self.TOP_LEVEL, key.replace(".", "__")] if x]
        ).upper()

    def _dot_alternative(self, key):
        return ".".join([self.TOP_LEVEL, key]) if self.TOP_LEVEL else key


class Agent(ConfigBuilder):
    TOP_LEVEL = "agent"

    def __init__(self):
        super().__init__()

        self.default_options = [
            # Some logger default values are handled by the logger
            ConfigOption(
                canonical_name="logger.level",
                default_value="",
                type_cast=str,
            ),
            ConfigOption(canonical_name="logger.path", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="logger.progname",
                default_value=DEFAULT_PROGNAME,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="security_logger.path",
                default_value=DEFAULT_SECURITY_LOG_PATH,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="security_logger.level",
                default_value=DEFAULT_SECURITY_LOG_LEVEL,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="python.rewrite",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="python.disable_policy_rewrites",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="python.enable_automatic_middleware",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="python.enable_drf_response_analysis",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="python.enable_rep",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="python.enable_profiler",
                default_value=False,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="python.assess.use_pure_python_hooks",
                default_value=False,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="polling.app_activity_ms",
                default_value=30000,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="polling.server_settings_ms",
                default_value=30_000,
                type_cast=int,
            ),
        ]


class Api(ConfigBuilder):
    TOP_LEVEL = "api"

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(canonical_name="url", default_value="", type_cast=str),
            ConfigOption(canonical_name="service_key", default_value="", type_cast=str),
            ConfigOption(canonical_name="api_key", default_value="", type_cast=str),
            ConfigOption(canonical_name="user_name", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="request_audit.enable",
                default_value=False,
                type_cast=bool,
            ),
            ConfigOption(
                canonical_name="request_audit.path",
                default_value=AGENT_CURR_WORKING_DIR,
                type_cast=str,
            ),
            ConfigOption(
                canonical_name="request_audit.requests",
                default_value=False,
                type_cast=bool,
            ),
            ConfigOption(
                canonical_name="request_audit.responses",
                default_value=False,
                type_cast=bool,
            ),
            ConfigOption(
                canonical_name="certificate.enable", default_value=False, type_cast=bool
            ),
            ConfigOption(
                canonical_name="certificate.ignore_cert_errors",
                default_value=False,
                type_cast=bool,
            ),
            ConfigOption(
                canonical_name="certificate.ca_file", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="certificate.cert_file", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="certificate.key_file", default_value="", type_cast=str
            ),
            ConfigOption(
                canonical_name="proxy.enable", default_value=False, type_cast=bool
            ),
            ConfigOption(canonical_name="proxy.url", default_value="", type_cast=str),
        ]


class Application(ConfigBuilder):
    TOP_LEVEL = "application"

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(canonical_name="code", default_value="", type_cast=str),
            ConfigOption(canonical_name="group", default_value="", type_cast=str),
            ConfigOption(canonical_name="metadata", default_value="", type_cast=str),
            ConfigOption(
                # TODO: PYT-2852 Revisit application name detection
                canonical_name="name",
                default_value=os.path.basename(AGENT_CURR_WORKING_DIR),
                type_cast=str,
            ),
            ConfigOption(canonical_name="path", default_value="", type_cast=str),
            ConfigOption(canonical_name="tags", default_value="", type_cast=str),
            ConfigOption(canonical_name="version", default_value="", type_cast=str),
            ConfigOption(canonical_name="session_id", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="session_metadata", default_value="", type_cast=str
            ),
        ]


class Assess(ConfigBuilder):
    TOP_LEVEL = "assess"

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="enable", default_value=None, type_cast=utils.str_to_bool
            ),
            ConfigOption(
                canonical_name="enable_scan_response",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="sampling.enable",
                default_value=False,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="sampling.baseline", default_value=5, type_cast=int
            ),
            ConfigOption(
                canonical_name="sampling.request_frequency",
                default_value=10,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="sampling.window_ms",
                default_value=180_000,
                type_cast=int,
            ),
            ConfigOption(canonical_name="tags", default_value="", type_cast=str),
            ConfigOption(
                canonical_name="rules.disabled_rules",
                default_value=[],
                type_cast=utils.parse_disabled_rules,
            ),
            ConfigOption(
                canonical_name="stacktraces",
                default_value="ALL",
                type_cast=utils.parse_stacktraces_options,
            ),
            ConfigOption(
                canonical_name="max_context_source_events",
                default_value=100,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="max_propagation_events",
                default_value=1000,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="time_limit_threshold",
                default_value=300000,  # 5 minutes in ms
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="max_rule_reported",
                default_value=100,
                type_cast=int,
            ),
        ]


class Inventory(ConfigBuilder):
    TOP_LEVEL = "inventory"

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="analyze_libraries",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="enable", default_value=True, type_cast=utils.str_to_bool
            ),
            ConfigOption(canonical_name="tags", default_value="", type_cast=str),
        ]


class Protect(ConfigBuilder):
    TOP_LEVEL = "protect"

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="enable", default_value=None, type_cast=utils.str_to_bool
            ),
            ConfigOption(
                canonical_name="samples.probed", default_value=50, type_cast=int
            ),
            ConfigOption(
                canonical_name="samples.blocked", default_value=25, type_cast=int
            ),
            ConfigOption(
                canonical_name="samples.exploited", default_value=100, type_cast=int
            ),
            ConfigOption(
                canonical_name="samples.blocked_at_perimeter",
                default_value=25,
                type_cast=int,
            ),
            ConfigOption(
                canonical_name="rules.bot-blocker.enable",
                default_value=False,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.cmd-injection.detect_chained_commands",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.cmd-injection.detect_parameter_command_backdoors",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.cmd-injection.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.disabled_rules",
                default_value=[],
                type_cast=utils.parse_disabled_rules,
            ),
            ConfigOption(
                canonical_name="rules.method-tampering.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.nosql-injection.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.path-traversal.detect_common_file_exploits",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.path-traversal.detect_custom_code_accessing_system_files",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.path-traversal.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.reflected-xss.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.sql-injection.detect_chained_queries",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.sql-injection.detect_dangerous_functions",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.sql-injection.detect_suspicious_unions",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.sql-injection.detect_tautologies",
                default_value=True,
                type_cast=utils.str_to_bool,
            ),
            ConfigOption(
                canonical_name="rules.sql-injection.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.ssrf.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.unsafe-file-upload.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
            ConfigOption(
                canonical_name="rules.xxe.mode",
                default_value=None,
                type_cast=utils.str_to_protect_mode_enum,
            ),
        ]


class Server(ConfigBuilder):
    TOP_LEVEL = "server"

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="name", default_value=utils.get_hostname(), type_cast=str
            ),
            ConfigOption(canonical_name="path", default_value="/", type_cast=str),
            ConfigOption(canonical_name="type", default_value="", type_cast=str),
            ConfigOption(canonical_name="version", default_value="", type_cast=str),
            ConfigOption(canonical_name="environment", default_value="", type_cast=str),
            ConfigOption(canonical_name="tags", default_value="", type_cast=str),
        ]


class Root(ConfigBuilder):
    TOP_LEVEL = None

    def __init__(self):
        super().__init__()

        self.default_options = [
            ConfigOption(
                canonical_name="enable", default_value=True, type_cast=utils.str_to_bool
            )
        ]


class SamplingLockConfig(Singleton):
    def init(self):
        self.lock_enable = False
        self.lock_baseline = False
        self.lock_frequency = False
        self.lock_window = False

    def set_sampling_config(self, value):
        if value == "enable":
            self.lock_enable = True
        if value == "baseline":
            self.lock_baseline = True
        if value == "request_frequency":
            self.lock_frequency = True
        if value == "window_ms":
            self.lock_window = True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


DISALLOWED_TAGS = [
    "CUSTOM_VALIDATED",
]


prompt_injection_triggers = [
    {
        "module": "openai",
        "class_name": "ChatCompletion",
        "method_name": ["create", "acreate"],
        "action": "OPENAI",
        "source": "KWARG:messages",
    },
    {
        "module": "openai",
        "class_name": "Completion",
        "method_name": ["create", "acreate"],
        "action": "OPENAI",
        "source": "KWARG:prompt",
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "prompt-injection",
        prompt_injection_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

def in_contrast_or_propagation_scope() -> bool: ...
def has_funchook() -> bool: ...
def get_current_scope() -> tuple: ...
def set_exact_scope(tuple) -> None: ...



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Iterable

try:
    from django.http.response import HttpResponseBase

    django_response_types = (HttpResponseBase,)
except ImportError:  # pragma: no cover
    django_response_types = ()

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

string_types = (str, bytes, bytearray)

list_iterator = type(iter(list()))
dict_iterator = type(iter(dict()))


def is_iterable(value):
    """
    :param value: any type
    :return: True if iterable type like list, dict but NOT string
             False if string or any other type like database collection types
                   (pymongo Collection implements iter but is not an iterable

    An iterable should meet one of the following conditions:
        * Has both __next__ and __iter__ methods, or
        * Has __getitem__ method

    For the purposes of Assess, we want to exclude stringy types from iteration.

    See https://stackoverflow.com/a/61139278 and associated question for
    relevant discussion.
    """
    return not isinstance(value, string_types) and (
        hasattr(value, "__getitem__") or isinstance(value, Iterable)
    )


def len_or_zero(value):
    try:
        return len(value)
    except Exception:
        return 0


def safe_iterator(it):
    if isinstance(it, (list_iterator, dict_iterator)):
        logger.debug(
            "WARNING: skipping iteration of non-seekable iterator: %s", type(it)
        )
        return

    if isinstance(it, (str, list, dict, tuple)) or (
        isinstance(it, django_response_types)
        and not safe_getattr(it, "streaming", True)
    ):
        for x in it:
            yield x
        return

    if not (hasattr(it, "tell") and hasattr(it, "seek")):
        logger.debug("WARNING: skipping iteration of non-seekable object: %s", type(it))
        return

    try:
        orig_pos = it.tell()
    except Exception:
        logger.debug(
            "WARNING: skipping iteration of non-tell()-able object: %s", type(it)
        )
        return

    try:
        for x in it:
            yield x
    except Exception:
        logger.debug("safe_iterator failed to iterate over %s", type(it))
    finally:
        it.seek(orig_pos)


def safe_getattr(obj, attr, default=None):
    """
    A getattr implementation that returns a default even if an exception is raised

    Some classes may override __getattribute__ to raise exceptions when certain
    attributes are accessed. When we are dealing with objects from the outside world,
    we want to be careful in our attribute access to avoid such exceptions and provide
    a safe default.
    """
    try:
        return getattr(obj, attr, default)
    except Exception:
        return default


def safe_getattr_list(obj, attr_names, default):
    """
    Try getattr for a list of several attribute names, returning the first one found. If
    none of the provided attributes exist on the object, return the default instead.
    """

    class NoAttr:
        pass

    for attr_name in attr_names:
        value = safe_getattr(obj, attr_name, NoAttr)
        if value is not NoAttr:
            return value
    return default



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.policy import constants
from contrast.agent.assess.assess_exceptions import ContrastAssessException
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction
from contrast.agent.assess.policy.trigger_actions import (
    redos_action,
    ssrf_action,
    fromstring_action,
    subprocess_action,
    openai_action,
)
from contrast.agent.policy.policy_node import PolicyNode
from contrast.api.trace_event import TraceEvent


class TriggerNode(PolicyNode):
    TRIGGER_ACTIONS = {
        constants.TRIGGER_ACTION_DEFAULT: DefaultAction(),
        constants.TRIGGER_ACTION_REDOS: redos_action.RedosAction(),
        constants.TRIGGER_ACTION_SSRF: ssrf_action.SsrfAction(),
        constants.TRIGGER_ACTION_FROMSTRING: fromstring_action.FromstringAction(),
        constants.TRIGGER_ACTION_SUBPROCESS: subprocess_action.SubprocessAction(),
        constants.TRIGGER_ACTION_OPENAI: openai_action.OpenAIAction(),
    }

    def __init__(
        self,
        module,
        class_name,
        instance_method,
        method_name,
        source=None,
        dataflow=True,
        good_value=None,
        bad_value=None,
        action=None,
        policy_patch=True,
        rule=None,
        protect_mode=False,
        unsafe_default=False,
    ):
        super().__init__(
            module,
            class_name,
            instance_method,
            method_name,
            source,
            None,
            policy_patch=policy_patch,
        )

        self.dataflow = dataflow

        self.good_value = (
            re.compile(good_value, flags=re.IGNORECASE) if good_value else None
        )
        self.bad_value = (
            re.compile(bad_value, flags=re.IGNORECASE) if bad_value else None
        )
        self.action = action or constants.TRIGGER_ACTION_DEFAULT

        self.rule = rule

        self.protect_mode = protect_mode
        self.unsafe_default = unsafe_default

        self.validate()

    @property
    def node_type(self):
        return TraceEvent.EventType.METHOD

    @property
    def dataflow_rule(self):
        return self.dataflow

    def validate(self):
        super().validate()

        if not self.dataflow_rule:
            return

        if not (self.sources and len(self.sources) != 0):
            raise ContrastAssessException(
                f"Trigger {self.name} did not have a proper source. Unable to create."
            )

    def get_matching_sources(self, self_obj, ret, args, kwargs):
        sources = []

        for source in self.sources:
            if source == constants.OBJECT:
                sources.append(self_obj)
            elif source == constants.ALL_ARGS:
                sources.append(args)
            elif source == constants.ALL_KWARGS:
                sources.append(kwargs)
            elif source == constants.RETURN:
                sources.append(ret)
            elif isinstance(source, int) and len(args) > source:
                sources.append(args[source])
            elif kwargs and source in kwargs:
                sources.append(kwargs[source])
            elif (
                isinstance(source, str)  # only consider kwarg sources
                and self.unsafe_default
                and source not in sources
            ):
                # if the default argument for this source is unsafe and we landed
                # here, we should trigger a finding. To do so, we add `None`,
                # with the assumption that most triggers will use `None`
                # as a default, but this won't always be correct. We may add a
                # trigger that uses a boolean as a default.

                # TODO: PYT-1764 special machinery to know what the default really is
                sources.append(None)
        return sources

    def get_protect_sources(self, args, kwargs):
        self_obj = args[0] if self.instance_method else None
        args = args[1:] if self.instance_method else args
        return self.get_matching_sources(self_obj, None, args, kwargs)

    @property
    def trigger_action(self):
        return self.TRIGGER_ACTIONS.get(self.action)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.providers.hardcoded_key import HardcodedKey
from contrast.agent.assess.rules.providers.hardcoded_password import HardcodedPassword
from contrast.utils.patch_utils import get_loaded_modules

PROVIDER_CLASSES = [HardcodedKey, HardcodedPassword]


def enable_providers():
    """
    Providers are non-dataflow rules that analyze the contents of a module.
    """
    for _, module in get_loaded_modules(use_for_patching=True).items():
        for provider_cls in PROVIDER_CLASSES:
            provider = provider_cls()
            if provider.applies_to(module):
                provider.analyze(module)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule import ProtectionRule


class Xss(BaseRule):
    """
    Cross Site Scripting Protection rule
    Currently only a prefilter / block at perimeter rule
    """

    RULE_NAME = "reflected-xss"

    @property
    def mode(self):
        """
        Always block at perimeter
        """
        mode = self.mode_from_settings()

        return (
            mode
            if mode in [ProtectionRule.OFF, ProtectionRule.MONITOR]
            else ProtectionRule.BLOCK_AT_PERIMETER
        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Optional

import os
import logging as stdlib_logging

import contrast
from contrast_vendor import structlog as logging
from contrast.agent.validator import ValidationException

logger = logging.getLogger("contrast")

DEBUG_LEVEL = stdlib_logging.DEBUG


def _log_extra_safely(
    log_message: Optional[str],
    error: Exception,
    original_func,
    args,
    kwargs,
    log_level="debug",
):
    if not contrast.telemetry_disabled():
        TELEMETRY = contrast.TELEMETRY
    else:
        TELEMETRY = None

    try:
        full_msg = (
            (log_message or "Exception in Contrast machinery") + ": " + str(error)
        )
        getattr(logger, log_level)(full_msg, exc_info=error)
        logger.debug("wrapped function args: %s", args)
        logger.debug("wrapped function kwargs: %s", kwargs)

        if TELEMETRY is not None:
            TELEMETRY.report_error(
                error=error, original_func=original_func, message=full_msg
            )
    except ValidationException as val_ex:
        logger.debug(f"Cannot report error to Telemetry: {str(val_ex)}")
    except Exception:
        # For complete safety, we're not going to try to log the logging error
        # because sometimes the logging error is actually an error (such as a recursive
        # error) within the logging module itself!
        pass


def fail_loudly(log_message=None, log_level="exception", return_value=None):
    """
    Decorator that will run the decorated function/method and, if
    an exception is raised, return a safe value and log the error.

    Note that SecurityException will always be re-raised.

    :param log_message: message to log in case of failure
    :param log_level: level to log in case of failure
    :param return_value: safe value to return in case of failure
    :return: original func return or return_value
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except contrast.SecurityException:
                raise
            except Exception as ex:
                _log_extra_safely(
                    log_message, ex, original_func, args, kwargs, log_level
                )
                if os.environ.get("CONTRAST_TESTING"):
                    logger.warn(
                        "Re-raising exception in fail_loudly (CONTRAST_TESTING is set)"
                    )
                    raise

            return return_value

        return run_safely

    return wrap


def fail_quietly(log_message=None, return_value=None):
    """
    Similar to fail_loudly (see above)

    This decorator is intended to handle cases where an exception may occur but won't
    disrupt normal operation of the agent. This decorator should be used to protect
    against external exceptions we can't prevent but still want to handle.

    In these cases, we log an error message and the exception traceback, both at DEBUG
    level.
    """

    def wrap(original_func):
        def run_safely(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except contrast.SecurityException:
                raise
            except Exception as ex:
                _log_extra_safely(log_message, ex, original_func, args, kwargs)
                if os.environ.get("CONTRAST_TESTING"):
                    logger.warn(
                        "Re-raising exception in fail_quietly (CONTRAST_TESTING is set)"
                    )
                    raise
            return return_value

        return run_safely

    return wrap


class cached_property(object):
    """
    https://github.com/pydanny/cached-property
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        if obj is None:
            return self

        value = obj.__dict__[self.func.__name__] = self.func(obj)
        return value



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.providers.hardcoded_value_rule import (
    HardcodedValueRule,
)


class HardcodedKey(HardcodedValueRule):
    # These are names, determined by the security team, that indicate a field
    # is likely to be a password or secret token of some sort.
    KEY_FIELD_NAMES = ["KEY", "AES", "DES", "IV", "SECRET"]

    # These are markers whose presence indicates that a field is more
    # likely to be a descriptor or requirement than an actual key.
    # We should ignore fields that contain them.
    NON_KEY_NAMES = ["CONTENT_CODES", "RESPONSE_CODES", "DIV"]

    @property
    def name(self):
        return "hardcoded-key"

    def is_name_valid(self, contant):
        name_in_key_fields = not any(self.fuzzy_match(contant, self.KEY_FIELD_NAMES))
        return name_in_key_fields or contant in self.NON_KEY_NAMES



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Database adapter patch for psycopg2.
This module's cursor doesn't have an `executescript` method.
"""

import os
from contrast_vendor.wrapt import register_post_import_hook

from contrast.patches.databases import dbapi2
from contrast.utils.decorators import fail_quietly

PSYCOPG2 = "psycopg2"
VENDOR = "PostgreSQL"


class Psycopg2Patcher(dbapi2.Dbapi2Patcher):
    @fail_quietly("failed to get database inventory information")
    def init_dbinfo(self, connection, connect_args, connect_kwargs):
        """
        Record DB inventory for a Postgres connection.

        Here we make a good effort to find connection params. There are several ways
        that these can be set, in the following order of priority (using dbname as an
        example):
        - using the `connection_factory` kwarg
        - as a kwarg itself - `dbname` or the deprecated `database`
        - via the dbname parameter in the dsn string
        - with the PGDATABASE environment variable

        Newer versions of psycopg2 (v2.7, ~2017) support connection.get_dsn_parameters,
        which provides a dictionary of the parsed connection params - we're interested
        in `dbname`.

        For now, it's still possible for us to miss the dbname (i.e. an old version of
        psycopg2 using the dsn string only), but this is unlikely and it would only
        affect inventory.
        """
        dsn_params = getattr(connection, "get_dsn_parameters", lambda: {})()
        host = (
            dsn_params.get("host")
            or connect_kwargs.get("host")
            or os.environ.get("PGHOST", "unknown_host")
        )
        port = (
            dsn_params.get("port")
            or connect_kwargs.get("port")
            or os.environ.get("PGPORT", "unknown_port")
        )
        dbname = (
            dsn_params.get("dbname")
            or connect_kwargs.get("dbname")
            or connect_kwargs.get("database")
            or os.environ.get("PGDATABASE", "unknown_database")
        )
        self.dbinfo["host"] = host
        self.dbinfo["port"] = port
        self.dbinfo["database"] = dbname


def instrument_psycopg2(psycopg2):
    dbapi2.instrument_adapter(
        psycopg2, VENDOR, Psycopg2Patcher, extra_cursors=[psycopg2.extensions.cursor]
    )


def register_patches():
    register_post_import_hook(instrument_psycopg2, PSYCOPG2)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Instrumentation for genshi globals function. Essentially what happens here is when
genshi calls the builtin eval function to eval the template code (and perform var lookup)
a custom mechanism is used in the genshi eval.py file. As a result we needed to patch
globals in order to inject our contrast__operator for modulo string operations.
"""

from contrast_vendor.wrapt import register_post_import_hook

from contrast.utils.patch_utils import build_and_apply_patch
from contrast.utils.patch_utils import wrap_and_watermark
from contrast_rewriter import populate_operator_module


GENSHI_TEMPALTE_ENGINE_EVAL_MODULE = "genshi.template.eval"


def build_genshi_eval_globals_patch(orig_func, _, rule_applicator):
    del rule_applicator

    def genshi_eval_globals_patch(wrapped, instance, args, kwargs):
        if len(args) > 0 and isinstance(args[0], dict):
            populate_operator_module(args[0])

        return orig_func(args)

    return wrap_and_watermark(orig_func, genshi_eval_globals_patch)


def patch_genshi_globals(genshi_eval_module):
    build_and_apply_patch(
        genshi_eval_module.LookupBase,
        "globals",
        build_genshi_eval_globals_patch,
        (None,),
    )


def register_patches():
    register_post_import_hook(patch_genshi_globals, GENSHI_TEMPALTE_ENGINE_EVAL_MODULE)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

"""
The sitecustomize.py file is automatically loaded by Python during interpreter
initialization. Our runner script ensures that it is on the PYTHONPATH which is
sufficient to make sure it is loaded.

See https://docs.python.org/3/library/site.html for additional details
"""

### This is the very first line of code we control in the app's python process ###

import contrast_rewriter

# NOTE: This must be applied prior to importing the agent itself. Do not import any
# other modules before registering the rewriter.
contrast_rewriter.register()


from contrast.agent.runner import start_runner

start_runner()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.assess.apply_trigger import cs__apply_trigger


def apply(
    rule, nodes, ret, orig_args, orig_kwargs=None, **kwargs
):  # pylint: disable=redefined-builtin
    """
    Iterates over all given trigger policy nodes and applies the given rule

    This method gets called from within a patched trigger function in order
    to determine whether the given rule has been violated.

    @param rule: `TriggerRule` instance representing rule to be evaluated
    @param nodes: List of `TriggerNode` instances
    @param ret: Result returned by the trigger function
    @param args: Tuple containing args passed to trigger function
    @param kwargs: Dict containing kwargs passed to trigger function
    """
    if not scope.in_trigger_scope():
        with scope.trigger_scope():
            orig_kwargs = orig_kwargs or {}

            context = contrast.CS__CONTEXT_TRACKER.current()
            if context is None:
                return

            for node in nodes:
                if node.instance_method:
                    self_obj = orig_args[0]
                    args = orig_args[1:]  # args[0] is `self` for instance methods
                else:
                    self_obj = None  # module-level functions do not have a self
                    args = orig_args

                possible_sources = node.get_matching_sources(
                    self_obj, ret, args, orig_kwargs
                )

                if not possible_sources:
                    if not rule.dataflow:
                        cs__apply_trigger(
                            context,
                            rule,
                            node,
                            None,
                            self_obj,
                            ret,
                            None,
                            args,
                            orig_kwargs,
                        )

                    return

                for source in possible_sources:
                    if node.trigger_action.unwind_source and isinstance(source, dict):
                        for key, value in source.items():
                            cs__apply_trigger(
                                context,
                                rule,
                                node,
                                key,
                                self_obj,
                                ret,
                                None,
                                args,
                                orig_kwargs,
                            )
                            # pass in the key here for building_finding
                            cs__apply_trigger(
                                context,
                                rule,
                                node,
                                value,
                                self_obj,
                                ret,
                                key,
                                args,
                                orig_kwargs,
                            )
                    elif node.trigger_action.unwind_source and isinstance(
                        source, (tuple, list)
                    ):
                        for item in source:
                            cs__apply_trigger(
                                context,
                                rule,
                                node,
                                item,
                                self_obj,
                                ret,
                                None,
                                args,
                                orig_kwargs,
                            )
                    else:
                        cs__apply_trigger(
                            context,
                            rule,
                            node,
                            source,
                            self_obj,
                            ret,
                            None,
                            args,
                            orig_kwargs,
                        )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.agent.middlewares.route_coverage.common import (
    build_route,
    DEFAULT_ROUTE_METHODS,
    get_normalized_uri,
    build_key,
)


def create_routes(app):
    """
    Returns all the routes registered to an app as a dict
    :param app: {Quart or Flask} app
    :return: dict {route_id:  api.Route}
    """
    routes = {}

    for rule in list(app.url_map.iter_rules()):
        view_func = app.view_functions[rule.endpoint]

        route = build_route(rule.endpoint, view_func)

        route_id = str(id(view_func))

        methods = rule.methods or DEFAULT_ROUTE_METHODS

        for method_type in methods:
            key = build_key(route_id, method_type)
            routes[key] = Route(
                verb=method_type,
                url=get_normalized_uri(str(rule)),
                route=route,
            )

    return routes


def get_view_func_for_request(request, app):
    """
    Find the view function for the current request in the app.
    :param request: current request
    :param app: {Quart or Flask} app
    :return: function
    """
    from werkzeug.exceptions import NotFound, MethodNotAllowed

    adapter = app.url_map.bind("empty")

    if None in (request, adapter):
        return None

    try:
        match = adapter.match(request.path, method=request.method)
    except (NotFound, MethodNotAllowed):
        match = None

    func = None
    if match is not None:
        func = app.view_functions[match[0]]

    return func



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import json

from contrast.agent.protect.rule.nosql_injection.mongo_nosql_scanner import (
    MongoNoSqlScanner,
)
from contrast.agent.protect.rule.sqli_rule import SqlInjection
from contrast.utils.decorators import fail_quietly


class NoSqlInjection(SqlInjection):
    """
    NoSQL Injection Protection rule
    """

    RULE_NAME = "nosql-injection"

    def get_database_scanner(self, database):
        return MongoNoSqlScanner()

    def skip_protect_analysis(self, user_input, args, kwargs):
        """
        nosql-injection has many potential user input types so
        let's not skip analysis
        """
        return False

    def convert_input(self, user_input):
        if not isinstance(user_input, (str, bytes)):
            user_input = obj_to_str(user_input)

        return super().convert_input(user_input)


@fail_quietly("Failed to convert nosql obj input to str", return_value="")
def obj_to_str(obj):
    """
    Convert one of the common obj types passed to pymongo methods into a string.

    :param obj: list, dict, bson or a type that inherits from collections.MutableMapping
    :return: str
    """
    try:
        return json.dumps(obj)
    except TypeError:
        # may encounter TypeError: Object of type ObjectId is not JSON serializable
        # so let's just make it a string
        return str(obj)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import re


import contrast
from contrast.agent import agent_lib
from contrast.agent.agent_lib import get_index_of_chained_cmd
from contrast.agent.agent_lib import does_command_contain_dangerous_path
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.cmd_injection.cmd_injection_rep_mixin import (
    CommandInjectionREPMixin,
)
from contrast.api.user_input import UserInput, DocumentType, InputType


class CmdInjection(BaseRule, CommandInjectionREPMixin):
    """
    Command Injection Protection rule
    """

    RULE_NAME = "cmd-injection"

    BIN_SH_C = "/bin/sh-c"

    CHAIN = ["&", ";", "|", ">", "<"]

    START_IDX = "start_idx"
    END_IDX = "end_idx"

    def find_attack(self, candidate_string=None, **kwargs):
        command_string = str(candidate_string) if candidate_string else None

        attack = super().find_attack(command_string, **kwargs)

        if not attack and command_string:
            evaluations_for_rule = self.evaluations_for_rule()
            return self.find_probable_attack(
                command_string, evaluations_for_rule, **kwargs
            )

        return attack

    def in_new_process(self):
        """
        Compare current pid to context pid
        """
        current_pid = os.getpid()

        original_pid = self.settings.pid

        return current_pid != original_pid

    def is_chained_command(self, command):
        """
        A command is chained if it is piped, backgrounded, or appended
        """
        return get_index_of_chained_cmd(command) != -1

    def does_contain_dangerous_path(self, command):
        return does_command_contain_dangerous_path(command) == 1

    def find_probable_attack(self, command_string, evaluations_for_rule, **kwargs):
        """
        Search through the attack string for a command that may have been executed
        """
        if not self.is_chained_command(
            command_string
        ) and not self.does_contain_dangerous_path(command_string):
            return None

        attack = None
        most_likely = None

        for evaluation in evaluations_for_rule:
            if not self.is_chained_command(
                evaluation.value
            ) and not self.does_contain_dangerous_path(evaluation.value):
                continue

            most_likely = evaluation
            break

        attack = self.build_attack_with_match(
            command_string, most_likely, attack, **kwargs
        )

        if not attack:
            return None

        self.log_rule_matched(most_likely, self.mode, command_string)

        return attack

    def build_sample(self, evaluation, command, **kwargs):
        sample = self.build_base_sample(evaluation)

        if command is not None:
            sample.details["command"] = command

        if self.START_IDX in kwargs or self.END_IDX in kwargs:
            sample.details["startIndex"] = kwargs.get(self.START_IDX, 0)
            sample.details["endIndex"] = kwargs.get(self.END_IDX, 0)
        elif command is not None:
            search_value = evaluation.value

            match = re.search(search_value, command, re.IGNORECASE)

            if match:
                sample.details["startIndex"] = match.start()
                sample.details["endIndex"] = match.end()

        return sample

    def infilter_kwargs(self, user_input, patch_policy):
        return dict(method=patch_policy.method_name, original_command=user_input)

    def skip_protect_analysis(self, user_input, args, kwargs):
        """
        cmdi rule supports list user input as well as str and bytes
        Do not skip protect analysis if user input is a  populated list
        """
        if isinstance(user_input, list) and user_input:
            return False

        return super().skip_protect_analysis(user_input, args, kwargs)

    def convert_input(self, user_input):
        if isinstance(user_input, list):
            user_input = " ".join(user_input)

        return super().convert_input(user_input)

    def infilter(self, match_string, **kwargs):
        context = contrast.CS__CONTEXT_TRACKER.current()

        if self.in_new_process():
            from contrast.agent.request_context import RequestContext

            context = RequestContext(context.environ)

        original_command = kwargs.get("original_command", "")

        if self.is_detect_parameter_command_backdoors_enabled:
            self.detect_command_backdoor(context, original_command)

        evaluations_for_rule = self.evaluations_for_rule()

        if not evaluations_for_rule:
            return

        if self.is_detect_chained_commands_enabled:
            evaluation = evaluations_for_rule[0]
            self.detect_command_chaining(original_command, evaluation)

        evaluation = evaluations_for_rule[0]
        self.detect_dangerous_path(original_command, evaluation)

        super().infilter(match_string, **kwargs)

    def detect_command_backdoor(self, context, command):
        """
        If we detect the user is supplying OS commands from a parameter
        then we'll block it. This is a common pattern from web shells
        and contrived applications.
        """
        parameter_key, parameter_value = self.find_matching_parameter(
            context, command
        ) or (None, None)
        if parameter_key and parameter_value:
            self.report_command_backdoor(command, parameter_key, parameter_value)

            raise contrast.SecurityException(rule_name=self.name)

    def detect_command_chaining(self, command, evaluation):
        """
        If we detected an attack inbound but we didn't see it in the command
        we still might want to fuzzy match on chained attacks.
        """
        index = get_index_of_chained_cmd(command)
        if index == -1:
            return

        self.report_chained_injection(evaluation, command, index)

        if self.is_blocked():
            raise contrast.SecurityException(rule_name=self.name)

    def detect_dangerous_path(self, command, evaluation):
        is_dangerous_path_detected = does_command_contain_dangerous_path(command)

        if is_dangerous_path_detected == 1:
            self.report_dangerous_path(evaluation, command)

            if self.is_blocked():
                raise contrast.SecurityException(rule_name=self.name)

    def find_matching_parameter(self, context, command):
        request = context.request

        if request and isinstance(command, str):
            normalized_command = self._normalize_str(command)

            parameters = request.params.dict_of_lists()

            for parameter_key, parameter_values in parameters.items():
                param = self._get_param_in_command(parameter_values, normalized_command)
                if param:
                    return parameter_key, param

        return None

    def _get_param_in_command(self, parameter_values, normalized_command):
        for param in parameter_values:
            if param and len(param) >= 2:
                normalized_param_value = self._normalize_str(param)

                if (
                    normalized_command == normalized_param_value
                    or normalized_command.startswith(self.BIN_SH_C)
                    or normalized_command.endswith(normalized_param_value)
                ):
                    return param

        return None

    def _normalize_str(self, string):
        return re.sub("\\s+", "", string).lower()

    def report_command_backdoor(self, command, parameter_key, parameter_value):
        evaluations_for_rule = self.evaluations_for_rule()
        if evaluations_for_rule:
            evaluation = evaluations_for_rule[0]
            evaluation.attack_count += 1
        else:
            evaluation = None

        sample = self.create_backdoor_command_sample(
            evaluation, command, parameter_key, parameter_value
        )

        attack = self.build_base_attack()

        attack.add_sample(sample)

        attack.set_response(self.response_from_mode(self.mode))

        self.log_rule_matched(evaluation, attack.response, parameter_value)

        self._append_to_context(attack)

    def report_chained_injection(self, evaluation, command, offset):
        start_idx = offset
        end_idx = len(command) - 1

        attack = self.build_or_append_attack(evaluation, candidate_string=command)

        result = self.build_attack_with_match(
            command, evaluation, attack, start_idx=start_idx, end_idx=end_idx
        )

        self._append_to_context(result)

    def report_dangerous_path(self, evaluation, command):
        attack = self.build_or_append_attack(evaluation, candidate_string=command)

        result = self.build_attack_with_match(command, evaluation, attack)

        self._append_to_context(result)

    def create_backdoor_command_sample(
        self, evaluation, command, parameter_key, parameter_value
    ):
        sample = self.build_base_sample(evaluation)
        user_input = UserInput(
            input_type=InputType.PARAMETER_VALUE,
            key=parameter_key,
            value=parameter_value,
            matcher_ids=[self.REP_DETECT_PARAMETER_COMMAND_BACKDOORS_ID],
            document_type=DocumentType.NORMAL,
        )
        sample.set_user_input(user_input)

        start_index = 0
        end_index = 0
        boundary_overrun_idx = 0
        input_boundary_idx = 0

        agent_lib_evaluation = agent_lib.check_cmd_injection_query(
            0, len(user_input.value), user_input.value
        )
        if bool(agent_lib_evaluation):
            start_index = agent_lib_evaluation.start_index
            end_index = agent_lib_evaluation.end_index
            boundary_overrun_idx = agent_lib_evaluation.boundary_overrun_index
            input_boundary_idx = agent_lib_evaluation.input_boundary_index

        sample.details["command"] = command
        sample.details["startIndex"] = start_index
        sample.details["endIndex"] = end_index
        sample.details["boundary_overrun_idx"] = boundary_overrun_idx
        sample.details["input_boundary_idx"] = input_boundary_idx

        return sample



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from sys import version_info

from contrast_vendor.isort import stdlibs


minor_version = version_info[1]
# TODO: PYT-3004 once isort supports 3.12 this special case should be removed
_version_string = f"py{version_info[0]}{11 if minor_version == 12 else minor_version}"
_stdlib_modules = getattr(stdlibs, _version_string).stdlib


def is_stdlib_module(module_name):
    """
    Returns True if module_name belongs to standard library module, False otherwise.

    NOTE: 'test' is included in _stdlib_modules so if we're testing this,
    we cannot pass in a module that starts with test.file...
    """
    top_module_name = module_name.split(".")[0]
    return top_module_name in _stdlib_modules



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import ast
from inspect import getsourcefile
from contrast.utils.decorators import fail_quietly
from contrast.utils.module_parser import get_ast_parsed_code


class CodeParserMixin(object):
    @fail_quietly("Unable to ast parse module")
    def find_hardcoded_str(self, cls_or_module, attribute_name, code_ast):
        """
        Checks the code ast to determine if the attribute has been
        assigned to a "true" hardcoded value (string, byte, etc)
        or if it has been assigned to a function that **evaluates**
        to a "true" hardcoded value.

        For reference:
                PRIVATE_KEY = "badbad" // is a real vulnerability

                PRIVATE_KEY = os.environ.get("secret") // not a vulnerability

        We use the suggested way to parse ast nodes, by implementing
        a NodeVisitor to look for assignment nodes that assign to
        string values and relying on ast.visit to do the node traversal.

        :param cls_or_module: class or module with code to parse
        :param attribute_name: attribute of the class or module to check for
        :param code_ast: ast tree representation of code in module.
            We return it and pass it back to this method for the next call
            to prevent parsing the module more than once.
        :return:
            1. bool determining if attribute_name is found to be assigned
                to a hardcoded str value
            2. the module str of code
        """
        if not code_ast:
            path = getsourcefile(cls_or_module)
            code_ast = get_ast_parsed_code(path)

        # must use a mutable type to be able to
        # update it in a lower scope. This is needed
        # because visit_Assign cannot return a value
        hardcoded_value = []

        class AssignToStr(ast.NodeVisitor):
            VALID_VALUE_TYPES = (
                ast.Str,
                ast.IfExp,
                ast.Bytes,
            )

            def visit_Assign(self, node):
                """
                An ast Assign node represents code that assigns values.
                This node has one or many targets but one value (what
                the targets get assigned to).
                https://greentreesnakes.readthedocs.io/en/latest/nodes.html#Assign

                We look if any of the targets have an id that matches
                the attribute_name If so, we check the node's value (which
                is equivalent to the attribute value) to see if it's a str type.
                """
                for target in node.targets:
                    if not isinstance(target, ast.Name):
                        # TODO: PYT-912 handle for assignment to class/instance
                        # such as self.key = ...
                        # which would be different target types
                        continue

                    if target.id == attribute_name:
                        if isinstance(node.value, self.VALID_VALUE_TYPES):
                            hardcoded_value.append(target)
                        break

        AssignToStr().visit(code_ast)

        hardcoded_found = bool(hardcoded_value)

        return hardcoded_found, code_ast



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.applies import apply_rule
from contrast.utils.patch_utils import add_watermark
from contrast.agent.policy import patch_manager

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def build_protect_patch(original_func, patch_policy):
    def protect_patch(*args, **kwargs):
        """
        Protect patch that will run in addition to running original_func.
        If we cannot run the protect rule, at the very least run the original_func.
        """
        return apply_rule(patch_policy, original_func, args, kwargs)

    return add_watermark(protect_patch)


def apply_protect_patch(patch_site, patch_policy):
    logger.debug("Applying protect patch to %s", patch_policy.name)

    original_func = getattr(patch_site, patch_policy.method_name)

    patch_manager.patch(
        patch_site,
        patch_policy.method_name,
        build_protect_patch(original_func, patch_policy),
    )



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.response.base_header_only_rule import (
    BaseHeaderOnlyRule,
)


class XXssProtectionDisabledRule(BaseHeaderOnlyRule):
    """
    Rule is violated if the X-XSS-Protection header is present and disables XSS
    protection. This is safe by default if the header isn't present.
    """

    @property
    def name(self):
        return "xxssprotection-header-disabled"

    @property
    def header_key(self):
        return "X-XSS-Protection"

    @property
    def good_value(self):
        return "1"

    @property
    def good_values(self):
        # unused
        pass

    def is_header_violated(self, headers):
        value = headers.get(self.header_key)
        if value is None:
            # no finding if header is missing (safe by default)
            return False, None

        if not value.startswith(self.good_value):
            properties = dict(
                type="Header",
                name=self.header_key,
                value=value,
            )
            return True, properties

        return False, None



#!/usr/bin/env python
import os
from glob import glob
from setuptools import setup, find_packages, Extension


extension_dir = "ext"
c_sources = glob(os.path.join(extension_dir, "*.c"))


extensions = [
    Extension(
        "hookspy.hookspy",
        c_sources,
        include_dirs=[extension_dir],
        extra_compile_args=[
            "-Wall",
            "-Wextra",
            "-Werror",
            "-Wno-unused-parameter",
            "-Wmissing-field-initializers",
        ],
    )
]


setup(
    name="hookspy",
    version="0.1.0",
    description="Find hook locations for python objects",
    url="",
    author="Dan D'Avella",
    author_email="daniel.davella@contrastsecurity.com",
    package_dir={"": "src"},
    packages=find_packages("src"),
    ext_modules=extensions,
    entry_points={
        "console_scripts": [
            "hookspy=hookspy:spy_main",
            "hookspy-autogen=hookspy:autogen_main",
        ]
    },
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.constants import (
    CURRENT_FINDING_VERSION,
    MINIMUM_FINDING_VERSION,
)


from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class Finding(object):
    def __init__(self, rule, properties, **kwargs):
        self.events = kwargs.get("events", [])
        self.rule_id = rule.name
        self.properties = properties
        self.hash_code = rule.generate_preflight_hash(**kwargs)
        self.preflight = ",".join([rule.name, self.hash_code])

        self.routes = []
        self.version = None
        logger.debug("Created finding for %s", rule.name)
        logger.debug("initial preflight value: %s", self.preflight)

    def set_version(self):
        """
        Determine the finding version depending on its attributes.
        """
        # dataflow or non-dataflow finding with route(s)
        if self.routes:
            self.version = CURRENT_FINDING_VERSION

        # non-dataflow rules without routes
        elif not self.events:
            self.version = CURRENT_FINDING_VERSION

        # dataflow finding without routes
        else:
            self.version = MINIMUM_FINDING_VERSION


def send_finding(finding, context=None):
    """
    Send a finding by either appending it to request context
    OR sending it immediately.

    If `context` exists, agent should not sent message immediately because
    current route needs to be appended in `append_route_to_findings`

    :param finding:  api.Finding instance
    :param context: Request context instance
    :return: None
    """
    if context:
        from contrast.agent.settings import Settings

        if (
            context.input_exclusions_trigger_time
            and Settings().exclusions.evaluate_assess_trigger_time_exclusions(
                context, finding
            )
        ):
            return

        context.findings.append(finding)
        return

    from contrast.reporting import ReportingClient
    from contrast.reporting import teamserver_messages

    ReportingClient().add_message(teamserver_messages.Preflight([finding], None))
